
from urllib.parse import urlparse

import urllib.request, urllib.parse, urllib.error

def upload(url, fp, background=False):
    raise Exception("Cannot upload to http store")

def view_url(url, expires=600, is_secure=False):
    return url

def exists(url):
    try:
        ifp = urllib.request.urlopen(url)
    except IOError:
        return False
    ifp.close()
    return True

def get_file(url, fp):
    ifp = urllib.request.urlopen(url)
    buf = ifp.read(10240)
    while buf:
        fp.write(buf)
        buf = ifp.read(10240)
    ifp.close()

def delete(url):
    pass # no-op

def netloc(url):
    u = urlparse(url)
    return u.netloc

def path(url):
    u = urlparse(url)
    return u.path



import traceback

class RenderError(Exception):
    def __init__(self, message="", stack=None, show_stack=False):
        self.message = message
        if stack and isinstance(stack, Exception):
            self.stack = getattr(stack, 'stack', None)
        elif stack and type(stack) in (str, str):
            self.stack = stack
        else:
            self.stack = None
        if not self.stack:
            self.stack = "".join(traceback.format_stack(None, 6)[:-1])

        if show_stack:
            self.message += '\n\tSTACK\t: %s' % self.stack.rstrip().replace("\n", "\n\t\t: ")
            
    def __str__(self):
        return "Render Error: %s" % (self.message)


class ValidateError(Exception):
    def __init__(self, message="", stack=None):
        self.message = message
        if stack and isinstance(stack, Exception):
            self.stack = getattr(stack, 'stack', None)
        elif stack and type(stack) in (str, str):
            self.stack = stack
        else:
            self.stack = None
        if not self.stack:
            self.stack = "".join(traceback.format_stack(None, 6)[:-1])

    def __str__(self):
        return "Validate Error: %s" % (self.message)

class CmdError(Exception):
    def __init__(self, retval=0, message="", cmd=None, stack=None):
        self.retval = retval
        self.message = message
        self.cmd = cmd
        if stack and isinstance(stack, Exception):
            self.stack = getattr(stack, 'stack', None)
        elif stack and type(stack) in (str, str):
            self.stack = stack
        else:
            self.stack = None
        if not self.stack:
            self.stack = "".join(traceback.format_stack(None, 6)[:-1])

    def __str__(self):
        if not self.cmd:
            cmdstr = ""
        elif type(self.cmd) in (list, tuple):
            cmdstr = "CMD: " + " ".join(((' ' in x and "'"+x+"'" or x) for x in self.cmd))
        else:
            cmdstr = "CMD: " + str(self.cmd)
        if self.message:
            message = "\n\tOUTPUT\t: %s" % self.message.strip().replace("\n","\n\t\t: ")
        else:
            message = ""
        return "EXIT: %d  %s%s" % (self.retval, cmdstr, message)



from objict import objict
from rest import settings
import requests

GEOIP_RESTIT_HOST = settings.get("GEOIP_RESTIT_HOST", None)


def getLocation(ip):
    """
    {
        "id": 14770,
        "created": 1653767808.0,
        "modified": 1671040424.0,
        "hostname": "ip70-187-221-89.oc.oc.cox.net",
        "ip": "70.187.221.89",
        "isp": "AS22773 Cox Communications Inc.",
        "provider": "ipinfo",
        "city": "Rancho Santa Margarita",
        "state": "California",
        "country": "US",
        "postal": None,
        "lat": 33.6409,
        "lng": -117.6031
    }
    """
    if GEOIP_RESTIT_HOST is None:
        return None
    params = {"ip": ip}
    url = f"https://{GEOIP_RESTIT_HOST}/rpc/location/ip/lookup"
    res = requests.get(url, params=params)
    if res.status_code != 200:
        return None
    data = objict.fromdict(res.json())
    data.provider = GEOIP_RESTIT_HOST
    return data.data



"""
This is where you can put handlers for running async background tasks

Task.Publish("myapp", "on_tq_test")
"""
# from datetime import datetime, timedelta
# from auditlog.models import PersistentLog
# from django.conf import settings
from incident.models import Event
from rest import helpers
from rest import settings


def new_event(task):
    data = task.data
    if "hostname" in data.metadata:
        data.hostname = data.metadata.hostname
    if "details" in data.metadata:
        data.details = data.metadata.details
    if "component" in data.metadata:
        data.component = data.metadata.component
    if "component_id" in data.metadata:
        data.component_id = data.metadata.component_id
    if "ip" in data.metadata:
        data.reporter_ip = data.metadata.ip
    Event.createFromDict(None, task.data)
    task.completed()


def firewall_block(task):
    # Task.Publish("incident", "firewall_block", {ip:"x.x.x.x"}, channel="tq_broadcast")
    helpers.blockIP(task.data.ip)
    task.log(f"{settings.HOSTNAME} - {task.data.ip} BLOCKED")
    task.completed()


def firewall_unblock(task):
    # Task.Publish("incident", "firewall_unblock", {ip:"x.x.x.x"}, channel="tq_broadcast")
    helpers.unblockIP(task.data.ip)
    task.log(f"{settings.HOSTNAME} - {task.data.ip} UNBLOCKED")
    task.completed()




from rest.decorators import periodic
from taskqueue.models import Task
from rest import settings


@periodic(hour=settings.get("AUDITLOG_CRON_HOUR", 9), minute=settings.get("AUDITLOG_CRON_MINUTE", 45))
def run_log_cleanup(force=False, verbose=False, now=None):
    # schedule pruning
    Task.Publish("auditlog", "on_cleanup", channel="tq_app_handler_cleanup")





from io import StringIO
import email
from email.utils import parseaddr, parsedate_to_datetime, getaddresses
from email.header import decode_header

from objict import objict


def parseRawMessage(msgobj):
    """
    Parse the email and return a dictionary of relevant data.
    """
    if isinstance(msgobj, str):
        msgobj = email.message_from_string(msgobj)

    subject = None
    message = None
    body = None
    html = None
    attachments = []
    body_parts = []
    html_parts = []

    if msgobj['Subject'] is not None:
        decodefrag = decode_header(msgobj['Subject'])
        subj_fragments = []
        for s, enc in decodefrag:
            if enc:
                s = str(s, enc).encode('utf8', 'replace')
            subj_fragments.append(s)
        subject = ''.join(subj_fragments)

    for part in msgobj.walk():
        attachment = parseAttachment(part)
        if attachment is None:
            continue
        if attachment.content:
            if attachment.content_type == "text/html":
                html_parts.append(attachment.content)
            elif attachment.content_type == "text/plain":
                body_parts.append(attachment.content)
        else:
            attachments.append(attachment)

    if len(body_parts):
        message = []
        body = "".join(body_parts)
        body = body.strip()
        # now lets parse the first part of the message that is not "quoted"
        blocks = 0
        for line in body.split('\n'):
            if line.startswith('>'):
                blocks += 1
                if blocks > 3:
                    break
                continue
            blocks = 0
            message.append(line.strip())
        message = "\n".join(message).strip()

    if len(html_parts):
        html = "".join(html_parts)

    from_addr = parseaddr(msgobj.get('From'))
    date_time = parsedate_to_datetime(msgobj.get('Date'))
    return objict({
        'subject': subject.strip(),
        'body': body,
        'sent_at': date_time,
        'message': message,
        'html': html,
        'from_email': from_addr[1],
        'from_name': from_addr[0],
        'to': msgobj.get("To"),
        'to_addrs': getaddresses(msgobj.get_all("To", [])),
        'cc': msgobj.get("Cc"),
        'cc_addrs': getaddresses(msgobj.get_all("Cc", [])),
        'attachments': attachments,
    })


def decodePayload(part):
    return str(part.get_payload(decode=True), part.get_content_charset(), 'replace')


def parseAttachment(message_part):
    content_disposition = message_part.get("Content-Disposition", None)
    if content_disposition is None:
        return None
    dispositions = content_disposition.strip().split(";")
    attachment = objict()
    attachment.dispositions = dispositions
    attachment.disposition = dispositions[0]
    attachment.payload = message_part.get_payload(decode=False)
    attachment.charset = message_part.get_content_charset()
    attachment.content_type = message_part.get_content_type()
    attachment.encoding = message_part.get("Content-Transfer-Encoding", "utf8")
    if attachment.disposition == "inline" and attachment.content_type in ["text/plain", "text/html"]:
        attachment.content = decodePayload(message_part)
    attachment.name = None
    attachment.create_date = None
    attachment.mod_date = None
    attachment.read_date = None
    # print dispositions
    for param in dispositions[1:]:
        name, value = param.split("=")
        name = name.strip().lower()
        if name == "filename":
            if value.startswith('"') or value.startswith("'"):
                value = value.strip()[1:-1]
            attachment.name = value
        elif name in ["create-date", "creation-date"]:
            attachment.create_date = value
        elif name == "modification-date":
            attachment.mod_date = value
        elif name == "read-date":
            attachment.read_date = value
    return attachment


def toFileObject(attachment):
    obj = StringIO(toString(attachment.payload))
    obj.name = attachment.name
    obj.size = len(attachment.payload)
    return obj


def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value




# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for OAuth.

Utilities for making it easier to work with OAuth 2.0
credentials.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import os
import stat
import threading

from .anyjson import simplejson
from .client import Storage as BaseStorage
from .client import Credentials


class CredentialsFileSymbolicLinkError(Exception):
  """Credentials files must not be symbolic links."""


class Storage(BaseStorage):
  """Store and retrieve a single credential to and from a file."""

  def __init__(self, filename):
    self._filename = filename
    self._lock = threading.Lock()

  def _validate_file(self):
    if os.path.islink(self._filename):
      raise CredentialsFileSymbolicLinkError(
          'File: %s is a symbolic link.' % self._filename)

  def acquire_lock(self):
    """Acquires any lock necessary to access this Storage.

    This lock is not reentrant."""
    self._lock.acquire()

  def release_lock(self):
    """Release the Storage lock.

    Trying to release a lock that isn't held will result in a
    RuntimeError.
    """
    self._lock.release()

  def locked_get(self):
    """Retrieve Credential from file.

    Returns:
      oauth2client.client.Credentials

    Raises:
      CredentialsFileSymbolicLinkError if the file is a symbolic link.
    """
    credentials = None
    self._validate_file()
    try:
      f = open(self._filename, 'rb')
      content = f.read()
      f.close()
    except IOError:
      return credentials

    try:
      credentials = Credentials.new_from_json(content)
      credentials.set_store(self)
    except ValueError:
      pass

    return credentials

  def _create_file_if_needed(self):
    """Create an empty file if necessary.

    This method will not initialize the file. Instead it implements a
    simple version of "touch" to ensure the file has been created.
    """
    if not os.path.exists(self._filename):
      old_umask = os.umask(0o177)
      try:
        open(self._filename, 'a+b').close()
      finally:
        os.umask(old_umask)

  def locked_put(self, credentials):
    """Write Credentials to file.

    Args:
      credentials: Credentials, the credentials to store.

    Raises:
      CredentialsFileSymbolicLinkError if the file is a symbolic link.
    """

    self._create_file_if_needed()
    self._validate_file()
    f = open(self._filename, 'wb')
    f.write(credentials.to_json())
    f.close()

  def locked_delete(self):
    """Delete Credentials file.

    Args:
      credentials: Credentials, the credentials to store.
    """

    os.unlink(self._filename)



from objict import objict
from datetime import datetime
import re
from .. import models as am
from location.models import GeoIP

IGNORE_RULES = [
    "100020"
]

LEVEL_REMAP_BY_RULE = {
    5402: 7,
    5710: 5
}


def removeNonAscii(input_str):
    """Remove all non-ASCII characters and escaped byte sequences from the input string."""
    # Remove escaped byte sequences
    cleaned_str = re.sub(r'\\x[0-9a-fA-F]{2}', '', input_str)
    # Remove non-ASCII characters
    return ''.join(char for char in cleaned_str if 32 <= ord(char) < 128)


def parseAlert(request, data):
    # helpers.log_print(data)
    try:
        data = objict.fromJSON(data.replace('\n', '\\n'))
    except Exception:
        data = objict.fromJSON(removeNonAscii(data))
    for key in data:
        data[key] = data[key].strip()

    if data.rule_id in IGNORE_RULES:
        return None
    if "test" in data.hostname and data.rule_id == "533":
        # bug on test ossec falsely report 533 events
        return None
    if data.rule_id == "510" and "/dev/.mount/utab" in data.text:
        return None
    # we care not for this field for now
    data.pop("logfile", None)
    if not data.text:
        raise Exception("invalid or missing json")
    alert = am.ServerOssecAlert(**data)
    alert.when = datetime.utcfromtimestamp(int(data.alert_id[:data.alert_id.find(".")]))
    # now lets parse the title
    title = alert.text[alert.text.find("Rule:") + 5:]
    # level to int
    level = title[title.find('(level') + 7:]
    alert.level = int(level[:level.find(')')].strip())
    title = title[:title.find('\n')].strip()
    pos = title.find("->")
    if pos > 0:
        title = title[pos + 2:]
    alert.title = title
    if alert.title.startswith("'"):
        alert.title = alert.title[1:-1]

    if data.hostname == "test":
        if data.rule_id == "31120" or "Web server" in title:
            return None

    # helpers.log_print(title, alert.title)
    # source ip (normally public ip of host)
    pos = alert.text.find("Src IP:")
    if pos > 1:
        src_ip = alert.text[alert.text.find("Src IP:") + 7:]
        alert.src_ip = src_ip[:src_ip.find('\n')].strip()

    irule = int(alert.rule_id)
    if irule == 5710:
        m = re.search(r"Invalid user (\S+) from (\S+)", data.text)
        if m and m.groups():
            alert.username = m.group(1)
            alert.src_ip = m.group(2).strip()
            alert.title = f"Attempt to login with invalid user: {alert.username}"
        else:
            m = re.search(r"Invalid user  from (\S+)", data.text)
            if m and m.groups():
                alert.username = "(empty string)"
                alert.src_ip = m.group(1).strip()
    elif irule == 2932:
        m = re.search(r"Installed: (\S+)", data.text)
        if m and m.groups():
            package = m.group(1)
            alert.title = "Yum Package Installed: {}".format(package)
    elif irule == 551:
        # Integrity checksum changed for: '/etc/ld.so.cache'
        m = re.search(r"Integrity checksum changed for: '(\S+)'", data.text)
        if m and m.groups():
            action = m.group(1)
            alert.title = "File Changed: {}".format(action)
    elif irule == 5715:
        m = re.search(r"Accepted publickey for (\S+).*ssh2: ([^\n\r]*)", data.text)
        if m and m.groups():
            ssh_sig = m.group(2)
            if " " in ssh_sig:
                kind, ssh_sig = ssh_sig.split(' ')
            alert.level = 8
            alert.username = m.group(1)
            alert.ssh_sig = ssh_sig
            alert.ssh_king = kind
            alert.title = f"SSH LOGIN:{alert.username}@{alert.hostname} from {alert.src_ip}"
            # member = findUserBySshSig(ssh_sig)
            # if member:
            #     alert.title = "SSH LOGIN user: {}".format(member.username)
    elif irule == 5501 or irule == 5502:
        # pam_unix(sshd:session): session opened for user git by (uid=0)
        m = re.search(r"session (\S+) for user (\S+)*", data.text)
        if m and m.groups():
            alert.action = m.group(1)
            alert.username = m.group(2)
            alert.title = f"session {alert.action} for user {alert.username}"
    elif irule == 5402:
        # TTY=pts/0 ; PWD=/opt/mm_protector ; USER=root ; COMMAND=/sbin/iptables -F
        m = re.search(r"sudo:\s*(\S+).*COMMAND=([^\n\r]*)", data.text)
        if m and m.groups():
            alert.username = m.group(1)
            alert.title = "sudo {}".format(m.group(2))
        alert.level = 7
    elif irule == 5706:
        m = re.search(r"identification string from (\S+) port (\S+)", data.text)
        if m and m.groups():
            alert.src_ip = m.group(1)
    elif irule == 5702:
        m = re.search(r"getaddrinfo for (\S+)", data.text)
        if m and m.groups():
            remote_host = m.group(1)
            alert.title = f"Reverse lookup failed for '{remote_host}'"
    elif irule == 554:
        m = re.search(r"New file '(\S+)' added", data.text)
        if m and m.groups():
            remote_file = m.group(1)
            alert.title = f"New file detected: '{remote_file}'"    
    elif irule == 100020:
        m = re.search(r"\[(\S+)\]", data.text)
        if m and m.groups():
            alert.src_ip = m.group(1)

    if alert.src_ip != None and len(alert.src_ip) > 6:
        # lets do a lookup for the src
        alert.geoip = GeoIP.lookup(alert.src_ip)
    # finally here we change the alert level
    if irule in LEVEL_REMAP_BY_RULE:
        alert.level = LEVEL_REMAP_BY_RULE[irule]
    if alert.title[0] in ["'", '"']:
        alert.title = alert.title[1:-1]
    if len(alert.title) > 80:
        alert.title = alert.title[:80] + "..."
    alert.save()
    return alert




from objict import objict
from io import StringIO
from django.http import StreamingHttpResponse
import csv


def flattenObject(obj, field_names):
    row = []
    for f in field_names:
        d = ""
        if "__" in f:
            f = f.replace("__", ".")
        if "." in f:
            # we could use obj.getFieldValue
            keys = f.split('.')
            f1 = keys[0]
            f2 = keys[1]
            if f1 == "metadata" and hasattr(obj, "getProperty"): 
                if len(keys) > 2:
                    d = obj.getProperty(f2, category=keys[2])
                else:
                    d = obj.getProperty(f2)
            elif hasattr(obj, "getFieldValue"):
                d = obj.getFieldValue(f, "t/a")
                if callable(d):
                    d = d()
            else:
                d1 = getattr(obj, f1, None)
                if d1 is not None:
                    if not hasattr(d1, f2):
                        if hasattr(d1, "first"):
                            d1 = d1.first()
                    d = getattr(d1, f2, "")
                    if callable(d):
                        d = d()                
        else:
            d = getattr(obj, f, "n/a")
            if callable(d):
                d = d()
        if d is None:
            d = "n/a"
        elif hasattr(d, "pk"):
            d = d.pk
        row.append(str(d))
    return row


def extractFieldNames(fields):
    header = []
    field_names = []
    for f in fields:
        if type(f) is tuple:
            r, f = f
            field_names.append(r)
        else:
            field_names.append(f)
        header.append(f)
    return header, field_names


def generateCSV(qset, fields, name, header_cols=None, values_list=False, output=None, stream=False):
    a = objict()
    a.name = name
    a.file = StringIO()
    if output:
        a.file = output
    csvwriter = csv.writer(a.file)
    header, field_names = extractFieldNames(fields)
    if header_cols:
        header = header_cols
    csvwriter.writerow(header)
    if values_list:
        for row in qset.values_list(*field_names):
            row = [str(x) for x in row]
            csvwriter.writerow(row)
    else:
        for obj in qset:
            csvwriter.writerow(flattenObject(obj, field_names))
    if hasattr(a.file, "getvalue"):
        a.data = a.file.getvalue()
    a.mimetype = "text/csv"
    return a


def iterCsvObject(items, writer, header, field_names):
    yield writer.writerow(header)
    for item in items:
        yield writer.writerow(flattenObject(item, field_names))


def generateCSVStream(qset, fields, name):
    import csv
    # check if we support stream mode
    header, field_names = extractFieldNames(fields)
    # rows = qset.values_list(*fields)
    pseudo_buffer = EchoWriter()
    writer = csv.writer(pseudo_buffer)
    return StreamingHttpResponse(
        iterCsvObject(qset, writer, header, field_names))


class EchoWriter(object):
    """An object that implements just the write method of the file-like
    interface.
    """
    def writeline(self, value):
        return "{}\n".format(value)

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value



from django.db import models
from location.models import GeoIP
from importlib import import_module
from django.conf import settings
from django.http import HttpRequest
from django.contrib.auth import login as auth_login, logout as auth_logout, authenticate
from datetime import datetime, timedelta
from rest import models as rm


def GetSessionByID(session_key):
    """
    Initialize same session as done for ``SessionMiddleware``.
    """
    engine = import_module(settings.SESSION_ENGINE)
    return engine.SessionStore(session_key)


class SessionLog(models.Model, rm.RestModel):
    """
    Logs session information
    """
    session_id = models.CharField(max_length=127, help_text="Django Session ID", db_index=True)
    user = models.ForeignKey("account.User", null=True, blank=True, related_name="sessions", on_delete=models.CASCADE)

    ip = models.CharField(max_length=127, null=True, blank=True, db_index=True)
    user_agent = models.TextField(help_text="Browser user agent", blank=True, null=True)
    created = models.DateTimeField(auto_now_add=True, editable=False, help_text="When session started", db_index=True)
    old_session = models.ForeignKey("self", help_text="Previous session of same user agent", related_name='next_session', null=True, blank=True, on_delete=models.CASCADE)
    location = models.ForeignKey(GeoIP, related_name="sessions", blank=True, null=True, default=None, on_delete=models.CASCADE)
    is_closed = models.BooleanField(default=False, blank=True, db_index=True)

    @property
    def session(self):
        return GetSessionByID(self.session_id)

    def __str__(self):
        return f"{self.user} - {self.ip} - {self.is_closed}"

    def isActive(self):
        return self.session.exists(self.session_id)

    def logout(self):
        session = self.session
        if session and session.exists(self.session_id):
            try:
                request = HttpRequest()
                request.session = session
                auth_logout(request)
            except Exception:
                pass
            session.delete()
        if not self.is_closed:
            self.is_closed = True
            self.save()

    def getLocation(self):
        if self.location is None and self.ip:
            self.location = GeoIP.get(self.ip)
            self.save()
        return self.location

    @staticmethod
    def GetSession(request):
        sid = request.session.session_key
        session = None
        if not sid:
            return None
        session = SessionLog.objects.filter(session_id=sid).last()
        if session is not None:
            return session
        user = None
        if request.user.is_authenticated:
            user = request.user
        session = SessionLog(
            session_id=sid,
            user=user,
            ip=request.ip,
            user_agent=request.META.get('HTTP_USER_AGENT', None))
        session.save()
        return session

    @staticmethod
    def GetExpired(days=90):
        stale = datetime.now() - timedelta(days=days)
        return SessionLog.objects.filter(user__isnull=False, is_closed=False, created__lte=stale)

    @staticmethod
    def LogOutExpired(days=90):
        qset = SessionLog.GetExpired(days)
        for slog in qset:
            slog.logout()

    @staticmethod
    def Clean(limit=1000):
        qset = SessionLog.objects.filter(user__isnull=False, is_closed=False).order_by("id")[:limit]
        for slog in qset:
            if not slog.isActive():
                slog.is_closed = True
                slog.save()

    def __unicode__(self):
        return "%s / %s@%s" % (self.session_id, self.user, self.ip)



from ..engines import ffmpeg
from ..engines import mp4box
from ..render_utils import *
from ..exceptions import *

import tempfile

PresetParameters =  {
    'audio_bitrate': {
        'description':	'Audio bitrate (kbps)',
        'required':	True,
        'kind':		'I',
        'configurable':	True,
    },
    'audio_channels': {
        'description':	'Audio channels',
        'required':	True,
        'kind':		'I',
        'configurable':	True,
    },
    'audio_freq': {
        'description':	'Audio frequency (hz)',
        'required':	True,
        'kind':		'I',
        'configurable':	True,
    },
    'video_bitrate': {
        'description':	'Video bitrate (kbps)',
        'required':	True,
        'kind':		'I',
        'configurable':	True,
    },
    'frame': {
        'description':	'Number of frames to encode',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'frame_rate': {
        'description':	'Frame rate',
        'required':	True,
        'kind':		'I',
        'configurable':	True,
    },
    'seek_time': {
        'description':	'Seek from start (sec)',
        'required':	True,
        'kind':		'F',
        'configurable':	True,
    },
    'pix_fmt': {
        'description':	'Pixel format',
        'required':	False,
        'kind':		'C',
        'configurable':	True,
        'choices':	'yuv420p,yuyv422,rgb24,bgr24,yuv422p,yuv444p,yuv410p,yuv411p,gray,monow,monob,yuvj420p,yuvj422p,yuvj444p,uyvy422,bgr8,bgr4,bgr4_byte,rgb8,rgb4,rgb4_byte,nv12,nv21,argb,rgba,abgr,bgra,gray16be,gray16le,yuv440p,yuvj440p,yuva420p,rgb48be,rgb48le,rgb565be,rgb565le,rgb555be,rgb555le,bgr565be,bgr565le,bgr555be,bgr555le,yuv420p16le,yuv420p16be,yuv422p16le,yuv422p16be,yuv444p16le,yuv444p16be,rgb444le,rgb444be,bgr444le,bgr444be,bgr48be,bgr48le,yuv420p9be,yuv420p9le,yuv420p10be,yuv420p10le,yuv422p10be,yuv422p10le,yuv444p9be,yuv444p9le,yuv444p10be,yuv444p10le,yuv422p9be,yuv422p9le',
        'value':	'yuv420p',
    },
    'key_frame': {
        'description':	'Key frame rate',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'duration': {
        'description':	'Duration',
        'required':	False,
        'kind':		'F',
        'configurable':	True,
    },
    'pass': {
        'description':	'Number of encoding passes',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
}

def render(item, renditiondef, params, cleanup=None, data=None):
    oname = params.get('stream_output')
    if not oname:
        fp = tempfile.NamedTemporaryFile(suffix=('.' + params.get('video_container', 'mp4')), prefix="t_ffmpeg_")
        oname = fp.name
        direct = False
    else:
        fp = None
        direct = True
    
    original = get_rendition(item, 'original')
    params['no_threads'] = False
    
    try:
        params = calc_size(params, original.width, original.height, rotate=original.get_meta('rotate'))
        
        if item.kind == 'L':
            ffmpeg.schedule(oname, original.filename, params, cleanup=cleanup, data=data)
        else:
            ffmpeg.run(oname, original.filename, params)

        # if fp and params.get('video_container', 'mp4') == 'mp4':
        # 	fp2 = tempfile.NamedTemporaryFile(suffix=('.' + params.get('video_container', 'mp4')), prefix="t_mp4box_")
        # 	mp4box.run(fp2.name, fp.name)
        # 	fp.close()
        # 	fp = fp2

        new_rendition(item, renditiondef, params.get('url') or fp or oname, kind='V', width=params.get('width', None), height=params.get('height', None), direct=direct)
    except (IOError, SyntaxError, ValueError, EOFError, AttributeError, IndexError, CmdError) as err:
        iserr = True
        raise RenderError('%s: %s' % (renditiondef.name, str(err)), stack=err, show_stack=True)
    else:
        iserr = False
    finally:
        if item.kind != 'L' and cleanup:
            cleanup(data, iserr)
        try:
            fp.close()
        except:
            pass



from rest import helpers
from rest import UberDict
from rest.requestex import RequestData
from django import http
from django.conf import settings
import time

from threading import currentThread
import types

_requests = {}


def get_request():
    i = currentThread().ident
    if i in _requests:
        return _requests[i]
    return None


class GlobalRequestMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.process_request(request)
        response = response or self.get_response(request)
        return response

    def _getReturnType(self, request):
        rtype = "text/plain"

        # figure return type
        if getattr(request, 'DATA', {}).get('_type', None):
            accept_list = [request.DATA.get('_type')]
        elif 'HTTP_ACCEPT' in getattr(request, 'META', {}):
            accept_list = request.META["HTTP_ACCEPT"].split(',')
        else:
            accept_list = []

        for i in range(len(accept_list)):
            a2 = accept_list[i].split(";", 1)
            if len(a2) == 2:
                accept_list[i] = a2[0]

        ext = "json"
        if 'application/json' in accept_list:
            rtype = "application/json"
        else:
            rtype = "text/plain"
        return rtype

    def process_exception(self, request, exception):
        try:
            rtype = self._getReturnType(request)
            payload = UberDict(error=str(exception) or getattr(type(exception), '__name__'), status=500)
            if type(exception) == http.Http404:
                payload.status = 404
            else:
                helpers.log_exception("process_exception")
            return http.HttpResponse(payload.toJSON(), status=200, content_type=rtype)
        except Exception:
            helpers.log_exception("another exception")
        return None

    def process_locale(self, request):
        locale = settings.DEFAULT_LANGUAGE
        if request.method == "POST":
            data = request.POST
        else:
            data = request.GET

        if "lang" in data:
            locale = data.get("lang")
        elif "locale" in data:
            locale = data.get("locale")
        elif 'HTTP_ACCEPT_LANGUAGE' in request.META:
            locale = request.META['HTTP_ACCEPT_LANGUAGE']

        lang = [x.strip()[:2] for x in locale.split(',')]
        if len(lang):
            request.LANGUAGE_CODE = lang[0]
        else:
            # we do this for when we get bad values above like empty strings
            request.LANGUAGE_CODE = settings.DEFAULT_LANGUAGE

        if "-" in request.LANGUAGE_CODE:
            request.COUNTRY_CODE = request.LANGUAGE_CODE.split('-')[0]
        else:
            request.COUNTRY_CODE = request.LANGUAGE_CODE

    def process_request(self, request):
        _requests[currentThread().ident] = request
        if not hasattr(request, "ip"):
            request.ip = helpers.getRemoteIP(request)
        request.setLogModel = types.MethodType(helpers.setLogModel, request)
        request._started = time.perf_counter()
        request._log_component = None
        request._log_pk = None
        request.logger = helpers.getLoggerByRequest(request)
        if not hasattr(request, "member"):
            request.member = None
        if not hasattr(request, "group"):
            request.group = None
        try:
            RequestData.upgradeRequest(request)
            if request.user.is_authenticated:
                if request.member is None:
                    request.member = request.user.getMember()
                if request.group is None:
                    gid = request.DATA.get(["group_id", "group"])
                    if gid:
                        request.group = request.member.getGroup(gid)
                if request.member is not None:
                    request.member.touchActivity()
        except Exception as err:
            helpers.log_exception("GlobalRequestMiddleware")
            print((str(err)))

        if settings.PROCESS_LOCALE:
            self.process_locale(request)



# Generated by Django 4.1.4 on 2023-08-22 19:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0008_incident_action_sent_incident_hostname_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='incident',
            name='reporter_ip',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=16, null=True),
        ),
    ]



from django.contrib import admin
from pushit.models import *

class ProductAdmin(admin.ModelAdmin):
    raw_id_fields = ['owner', 'group', 'library']
    search_fields = ["name"]
    list_display = ('name', 'kind', 'kind')

class ReleaseAdmin(admin.ModelAdmin):
    search_fields = ["name"]
    list_display = ('created', 'product', 'version_num', 'version_str')


admin.site.register(Product, ProductAdmin)
admin.site.register(Release, ReleaseAdmin)



from rest import decorators as rd
from rest import views as rv
from rest import helpers as rh
from rest import settings
from rest import crypto
from urllib.parse import urlencode
from django.shortcuts import redirect
from account.oauth import google
from account.models import Member
from objict import objict
from datetime import datetime, timedelta


@rd.url('oauth/google/login')
@rd.never_cache
def oauth_google_login(request):
    code = request.DATA.get("code")
    error = request.DATA.get("error", "unknown error")
    state = request.DATA.get("state")
    app_url = settings.DEFAULT_LOGIN_URL

    rh.log_print("google/login", request.DATA.toDict(), request.session.get("state"))

    if state:
        state = objict.fromJSON(rh.hexToString(state))
        rh.log_print("state", state)
        app_url = state.url

    if not code:
        params = urlencode({'error': error})
        return redirect(f"{app_url}?{params}")

    redirect_uri = f"{settings.BASE_URL_SECURE}rpc/account/oauth/google/login"
    auth_data = google.getAccessToken(code, redirect_uri)
    if auth_data is None or auth_data.access_token is None:
        params = urlencode({'error': "failed to get access token from google"})
        return redirect(f"{app_url}?{params}")

    user_data = google.getUserInfo(auth_data.access_token)
    if user_data is None:
        params = urlencode({'error': "failed to get user data from google"})
        return redirect(f"{app_url}?{params}")

    if not user_data.email:
        params = urlencode({'error': "no email with account"})
        return redirect(f"{app_url}?{params}")

    # TODO allow new accounts?
    member = Member.objects.filter(email=user_data.email).last()
    if member is None:
        params = urlencode({'error': "user not found"})
        return redirect(f"{app_url}?{params}")

    member.setProperties(auth_data, category="google_auth")
    member.setProperties(user_data, category="google")

    if not member.first_name and user_data.given_name:
        member.first_name = user_data.given_name

    if not member.last_name and user_data.family_name:
        member.last_name = user_data.family_name

    if not member.display_name and user_data.name:
        member.display_name = user_data.name

    member.auth_code = crypto.randomString(16)
    member.auth_code_expires = datetime.now() + timedelta(minutes=5)
    member.save()
    member.auditLog("user succesfully authenticated with google", "google_oauth", level=17)

    params = urlencode({'oauth_code': member.auth_code, "username":member.username})
    return redirect(f"{app_url}?{params}")



from ..exceptions import *
from ..render_utils import RenderProc, PipeThread

from django import db

import select
import fcntl
import os
import sys

class RtmpDumpRun(PipeThread):
    def __init__(self, *args, **kwargs):
        return super(RtmpDumpRun, self).__init__(*args, **kwargs)

    def __del__(self):
        os.close(self.args[1])
        super(RtmpDumpRun, self).__del__()

    def read(self, inf, outfd):
        data = inf.read(10240)
        return data

    def write(self, data, inf, outfd):
        return os.write(outfd, data)

def thread(url):
    from .rtmp import rtmp
    (rp, wp) = os.pipe()
    inf = rtmp.dump(str(url))
    RtmpDumpRun(inf, wp, daemon=True)
    return rp



from .models import *
from . import stores

from django.shortcuts import render, get_object_or_404, Http404
from django.http import HttpResponse

def smil(request, item_id, token=None):
    item = get_object_or_404(MediaItem, pk=item_id)
    #if not item.check_token(token):
    #   item.assertAcl("R", request.user)
    renditions = item.renditions.filter(use='video').order_by('-created')
    rseen = []
    ret = []
    base = ""
    for r in renditions:
        if r.rendition_definition_id in rseen:
            continue
        rseen.append(r.rendition_definition_id)
        urls = r.view_url().split("/")
        url = urls.pop()
        base = "/".join(urls)
        ret.append({
            'url': url,
            'bitrate': r.bitrate(),
        })

    return render(request, 'medialib/smil',
        RequestContext(request, {
            'base': base,
            'item': item,
            'renditions': ret,
        }),
        mimetype="application/smil",
    )


def hlsIndex(request, rendition_id, token=None):
    rendition = get_object_or_404(MediaItemRendition, pk=rendition_id)
    if not rendition.mediaitem.check_token(token):
        rendition.mediaitem.assertAcl("R", request.user)

    try:
        bufsize = int(request.DATA.get('buffer', 30))
    except ValueError:
        bufsize = 30

    segment_duration = rendition.get_meta('segment_duration', 1, int)
    final = rendition.get_meta('final')

    if final and not request.DATA.get('live', False):
        segments = rendition.segments.order_by('segment').values_list('segment', 'url', 'duration')
        if not segments:
            raise Http404
    else:
        segments = list(rendition.segments.order_by('-segment')[:int(bufsize/segment_duration)].values_list('segment', 'url', 'duration'))
        segments.reverse()

    try:
        outbuf = """#EXTM3U
#EXT-X-TARGETDURATION:%d
#EXT-X-MEDIA-SEQUENCE:%d
#EXT-X-SLIDINGWINDOW
""" % (segment_duration, segments[0][0]+3)
        for s in segments:
            outbuf += "#EXTINF:%d,\n" % s[2]
            outbuf += stores.view_url(s[1], is_secure=request.is_secure(), request=request) + "\n"
        if final:
            outbuf += "#EXT-X-ENDLIST\n"
    except IndexError:
        outbuf = """#EXTM3U
#EXT-X-TARGETDURATION:3
#EXT-X-MEDIA-SEQUENCE:0
#EXT-X-SLIDINGWINDOW
#EXTINF:3,
/static/video/wait.ts
"""
    return HttpResponse(outbuf, mimetype="application/vnd.apple.mpegurl")

def instances(request):
    if not request.user.is_superuser:
        raise NotAuthorized

    start = request.DATA.get('start', 0)
    count = request.DATA.get('count', 25)

    instances = RenderInstance.objects.all().order_by('-priority', '-shutdown', '-priority', 'started')[start:start+count]
    return render(request, 'medialib/instances.html', {
        'title': 'Rendering Instances',
        'instances': instances,
        'more': (len(instances) == count) and (start+count+1) or None
    })

def renderErrors(request):
    if not request.user.is_superuser:
        raise NotAuthorized

    start = request.DATA.get('start', 0)
    count = request.DATA.get('count', 25)

    items = MediaItem.objects.filter(state=110).order_by('-id')[start:start+count]
    return render(request, 'medialib/items.html', {
        'title': 'Rendering Errors',
        'items': items,
        'more': (len(items) == count) and (start+count+1) or None
    })

def renderPending(request):
    if not request.user.is_superuser:
        raise NotAuthorized

    items = MediaItem.objects.filter(state__gte=100, state__lt=110).order_by('id')
    return render(request, 'medialib/items.html', {
        'title': 'Pending Renders',
        'items': items,
    })

def showLibrary(request, library_id):
    if not request.user.is_superuser:
        raise NotAuthorized

    library = MediaLibrary.objects.get(pk=library_id)
    libraries = MediaLibrary.objects.all()
    return render(request, 'medialib/library.html', {
        'title': 'Pending Renders',
        'library': library,
        'libraries': libraries
    })

def showLibraryOld(request, library_id):
    if not request.user.is_superuser:
        raise NotAuthorized

    library = MediaLibrary.objects.get(pk=library_id)
    return render(request, 'medialib/testpicker.html', {
        'title': 'Pending Renders',
        'library': library,
    })



# Generated by Django 4.0.3 on 2022-04-04 23:52

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    dependencies = [
        ('location', '0001_initial'),
        ('account', '0006_member_security_token'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthSession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('last_activity', models.DateTimeField(db_index=True)),
                ('signature', models.CharField(db_index=True, max_length=127)),
                ('ip', models.CharField(blank=True, db_index=True, max_length=127, null=True)),
                ('user_agent', models.TextField(blank=True, null=True)),
                ('browser', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('os', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('device', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('location', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='auth_sessions', to='location.geoip')),
                ('member', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='auth_sessions', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



"""Channel notifications support.

Classes and functions to support channel subscriptions and notifications
on those channels.

Notes:
  - This code is based on experimental APIs and is subject to change.
  - Notification does not do deduplication of notification ids, that's up to
    the receiver.
  - Storing the Channel between calls is up to the caller.


Example setting up a channel:

  # Create a new channel that gets notifications via webhook.
  channel = new_webhook_channel("https://example.com/my_web_hook")

  # Store the channel, keyed by 'channel.id'. Store it before calling the
  # watch method because notifications may start arriving before the watch
  # method returns.
  ...

  resp = service.objects().watchAll(
    bucket="some_bucket_id", body=channel.body()).execute()
  channel.update(resp)

  # Store the channel, keyed by 'channel.id'. Store it after being updated
  # since the resource_id value will now be correct, and that's needed to
  # stop a subscription.
  ...


An example Webhook implementation using webapp2. Note that webapp2 puts
headers in a case insensitive dictionary, as headers aren't guaranteed to
always be upper case.

  id = self.request.headers[X_GOOG_CHANNEL_ID]

  # Retrieve the channel by id.
  channel = ...

  # Parse notification from the headers, including validating the id.
  n = notification_from_headers(channel, self.request.headers)

  # Do app specific stuff with the notification here.
  if n.resource_state == 'sync':
    # Code to handle sync state.
  elif n.resource_state == 'exists':
    # Code to handle the exists state.
  elif n.resource_state == 'not_exists':
    # Code to handle the not exists state.


Example of unsubscribing.

  service.channels().stop(channel.body())
"""

import datetime
import uuid

from apiclient import errors
from oauth2client import util


# The unix time epoch starts at midnight 1970.
EPOCH = datetime.datetime.utcfromtimestamp(0)

# Map the names of the parameters in the JSON channel description to
# the parameter names we use in the Channel class.
CHANNEL_PARAMS = {
    'address': 'address',
    'id': 'id',
    'expiration': 'expiration',
    'params': 'params',
    'resourceId': 'resource_id',
    'resourceUri': 'resource_uri',
    'type': 'type',
    'token': 'token',
    }

X_GOOG_CHANNEL_ID     = 'X-GOOG-CHANNEL-ID'
X_GOOG_MESSAGE_NUMBER = 'X-GOOG-MESSAGE-NUMBER'
X_GOOG_RESOURCE_STATE = 'X-GOOG-RESOURCE-STATE'
X_GOOG_RESOURCE_URI   = 'X-GOOG-RESOURCE-URI'
X_GOOG_RESOURCE_ID    = 'X-GOOG-RESOURCE-ID'


def _upper_header_keys(headers):
  new_headers = {}
  for k, v in list(headers.items()):
    new_headers[k.upper()] = v
  return new_headers


class Notification(object):
  """A Notification from a Channel.

  Notifications are not usually constructed directly, but are returned
  from functions like notification_from_headers().

  Attributes:
    message_number: int, The unique id number of this notification.
    state: str, The state of the resource being monitored.
    uri: str, The address of the resource being monitored.
    resource_id: str, The unique identifier of the version of the resource at
      this event.
  """
  @util.positional(5)
  def __init__(self, message_number, state, resource_uri, resource_id):
    """Notification constructor.

    Args:
      message_number: int, The unique id number of this notification.
      state: str, The state of the resource being monitored. Can be one
        of "exists", "not_exists", or "sync".
      resource_uri: str, The address of the resource being monitored.
      resource_id: str, The identifier of the watched resource.
    """
    self.message_number = message_number
    self.state = state
    self.resource_uri = resource_uri
    self.resource_id = resource_id


class Channel(object):
  """A Channel for notifications.

  Usually not constructed directly, instead it is returned from helper
  functions like new_webhook_channel().

  Attributes:
    type: str, The type of delivery mechanism used by this channel. For
      example, 'web_hook'.
    id: str, A UUID for the channel.
    token: str, An arbitrary string associated with the channel that
      is delivered to the target address with each event delivered
      over this channel.
    address: str, The address of the receiving entity where events are
      delivered. Specific to the channel type.
    expiration: int, The time, in milliseconds from the epoch, when this
      channel will expire.
    params: dict, A dictionary of string to string, with additional parameters
      controlling delivery channel behavior.
    resource_id: str, An opaque id that identifies the resource that is
      being watched. Stable across different API versions.
    resource_uri: str, The canonicalized ID of the watched resource.
  """

  @util.positional(5)
  def __init__(self, type, id, token, address, expiration=None,
               params=None, resource_id="", resource_uri=""):
    """Create a new Channel.

    In user code, this Channel constructor will not typically be called
    manually since there are functions for creating channels for each specific
    type with a more customized set of arguments to pass.

    Args:
      type: str, The type of delivery mechanism used by this channel. For
        example, 'web_hook'.
      id: str, A UUID for the channel.
      token: str, An arbitrary string associated with the channel that
        is delivered to the target address with each event delivered
        over this channel.
      address: str,  The address of the receiving entity where events are
        delivered. Specific to the channel type.
      expiration: int, The time, in milliseconds from the epoch, when this
        channel will expire.
      params: dict, A dictionary of string to string, with additional parameters
        controlling delivery channel behavior.
      resource_id: str, An opaque id that identifies the resource that is
        being watched. Stable across different API versions.
      resource_uri: str, The canonicalized ID of the watched resource.
    """
    self.type = type
    self.id = id
    self.token = token
    self.address = address
    self.expiration = expiration
    self.params = params
    self.resource_id = resource_id
    self.resource_uri = resource_uri

  def body(self):
    """Build a body from the Channel.

    Constructs a dictionary that's appropriate for passing into watch()
    methods as the value of body argument.

    Returns:
      A dictionary representation of the channel.
    """
    result = {
        'id': self.id,
        'token': self.token,
        'type': self.type,
        'address': self.address
        }
    if self.params:
      result['params'] = self.params
    if self.resource_id:
      result['resourceId'] = self.resource_id
    if self.resource_uri:
      result['resourceUri'] = self.resource_uri
    if self.expiration:
      result['expiration'] = self.expiration

    return result

  def update(self, resp):
    """Update a channel with information from the response of watch().

    When a request is sent to watch() a resource, the response returned
    from the watch() request is a dictionary with updated channel information,
    such as the resource_id, which is needed when stopping a subscription.

    Args:
      resp: dict, The response from a watch() method.
    """
    for json_name, param_name in list(CHANNEL_PARAMS.items()):
      value = resp.get(json_name)
      if value is not None:
        setattr(self, param_name, value)


def notification_from_headers(channel, headers):
  """Parse a notification from the webhook request headers, validate
    the notification, and return a Notification object.

  Args:
    channel: Channel, The channel that the notification is associated with.
    headers: dict, A dictionary like object that contains the request headers
      from the webhook HTTP request.

  Returns:
    A Notification object.

  Raises:
    errors.InvalidNotificationError if the notification is invalid.
    ValueError if the X-GOOG-MESSAGE-NUMBER can't be converted to an int.
  """
  headers = _upper_header_keys(headers)
  channel_id = headers[X_GOOG_CHANNEL_ID]
  if channel.id != channel_id:
    raise errors.InvalidNotificationError(
        'Channel id mismatch: %s != %s' % (channel.id, channel_id))
  else:
    message_number = int(headers[X_GOOG_MESSAGE_NUMBER])
    state = headers[X_GOOG_RESOURCE_STATE]
    resource_uri = headers[X_GOOG_RESOURCE_URI]
    resource_id = headers[X_GOOG_RESOURCE_ID]
    return Notification(message_number, state, resource_uri, resource_id)


@util.positional(2)
def new_webhook_channel(url, token=None, expiration=None, params=None):
    """Create a new webhook Channel.

    Args:
      url: str, URL to post notifications to.
      token: str, An arbitrary string associated with the channel that
        is delivered to the target address with each notification delivered
        over this channel.
      expiration: datetime.datetime, A time in the future when the channel
        should expire. Can also be None if the subscription should use the
        default expiration. Note that different services may have different
        limits on how long a subscription lasts. Check the response from the
        watch() method to see the value the service has set for an expiration
        time.
      params: dict, Extra parameters to pass on channel creation. Currently
        not used for webhook channels.
    """
    expiration_ms = 0
    if expiration:
      delta = expiration - EPOCH
      expiration_ms = delta.microseconds/1000 + (
          delta.seconds + delta.days*24*3600)*1000
      if expiration_ms < 0:
        expiration_ms = 0

    return Channel('web_hook', str(uuid.uuid4()),
                   token, url, expiration=expiration_ms,
                   params=params)




from django.db import models as dm
from rest import fields as rf
from rest import helpers as rh
from objict import objict


class JSONMetaData(dm.Model):
    class Meta:
        abstract = True
        
    metadata = rf.JSONField()

    def set_metadata(self, value):
        # override set_metadata to call this for updating metadata vs deleting metadata
        if self.metadata is None:
            self.metadata = objict()
        if isinstance(value, dict):
            value = objict.fromdict(value)
            rh.dictDeepUpdate(self.metadata, value)

    def getProperty(self, key, default=None):
        if self.metadata is None:
            return None
        if not isinstance(self.metadata, objict):
            self.metadata = objict.fromdict(self.metadata)
        return self.metadata.get(key, default)

    def setProperty(self, key, value):
        if self.metadata is None:
            self.metadata = objict()
        if not isinstance(self.metadata, objict):
            self.metadata = objict.fromdict(self.metadata)
        self.metadata.set(key, value)



# Generated by Django 4.1.1 on 2022-09-24 21:56

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0010_delete_authtoken'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('token', models.TextField(db_index=True, unique=True)),
                ('role', models.CharField(blank=True, default=None, max_length=128, null=True)),
                ('signature', models.CharField(blank=True, db_index=True, default=None, max_length=128, null=True)),
                ('member', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_tokens', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



from django.db import models
from rest import models as rm
from rest import settings
from datetime import datetime, timedelta
from .. import geolocate

GEOIP_LOOKUP_BY_SUBNET = settings.get("GEOIP_LOOKUP_BY_SUBNET", True)


class GeoIP(models.Model, rm.RestModel):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    hostname = models.CharField(max_length=255, blank=True, null=True, default=None)
    ip = models.CharField(max_length=64, db_index=True)
    subnet = models.CharField(max_length=64, db_index=True, default=None, null=True)
    isp = models.CharField(max_length=84)

    city = models.CharField(max_length=64, blank=True, null=True)
    state = models.CharField(max_length=64, blank=True, null=True)
    country = models.CharField(max_length=64, blank=True, null=True)
    postal = models.CharField(max_length=32, blank=True, null=True)

    lat = models.FloatField(default=0.0, blank=True)
    lng = models.FloatField(default=0.0, blank=True)

    def __str__(self):
        return f"<GeoIP: {self.ip} {self.lat}:{self.lng}"

    def isStale(self, days=300):
        stale = self.modified + timedelta(days=days)
        return stale < datetime.now()

    def refresh(self, using=None):
        if self.ip == "127.0.0.1":
            self.isp = "local"
            self.hostname = "localhost"
            self.subnet = "127.0.0"
            self.modified = datetime.now()
            self.save(using=using)
            return
        res = geolocate.locateByIP(self.ip)
        if res is None:
            # just save the
            if not self.pk:
                self.isp = "unknown"
            self.modified = datetime.now()
            self.save(using=using)
            return
        self.hostname = res.hostname
        if res.isp is None:
            self.isp = "unknown"
        else:
            self.isp = res.isp[:83]
        self.city = res.city
        self.state = res.state
        self.country = res.country
        self.postal = res.postal
        try:
            self.lat = float(res.latitude)
            self.lng = float(res.longitude)
        except Exception:
            pass
        self.save(using=using)

    @classmethod
    def get(cls, ip, force_refresh=False, stale_after=300):
        return cls.lookup(ip, force_refresh, stale_after)

    @classmethod
    def lookup(cls, ip, force_refresh=False, stale_after=300, using=None):
        if isinstance(ip, str):
            ip = ip.strip()
        if not geolocate.isIP(ip):
            ip = geolocate.dnsToIP(ip)
        subnet = ip[:ip.rfind(".")]
        gip = GeoIP.objects.filter(ip=ip).first()
        if gip is None:
            gip = GeoIP(ip=ip, subnet=subnet)
            if GEOIP_LOOKUP_BY_SUBNET:
                subgip = GeoIP.objects.filter(subnet=subnet).last()
                if subgip:
                    gip.lat = subgip.lat
                    gip.lng = subgip.lng
                    gip.city = subgip.city
                    gip.state = subgip.state
                    gip.country = subgip.country
                    gip.postal = subgip.postal
                    gip.isp = subgip.isp[:83]
                    gip.save()
                else:
                    gip.refresh(using=using)
            else:
                gip.refresh(using=using)
        else:
            if force_refresh or gip.isStale(stale_after):
                gip.refresh(using=using)
            elif gip.subnet is None:
                gip.subnet = subnet
                gip.save()
        return gip



from ..exceptions import *

import os
import tempfile
import subprocess
import re
from medialib import utils


ffmpeg_cmd = ['ffmpeg']
ffmpeg_args = []
ffmpeg_args_url = []
ffmpeg_args_file = ['-vcodec', 'copy', '-acodec', 'copy', '-f', 'null', '/dev/null']

def getinfo(infile, params={}):
    fp = tempfile.TemporaryFile()
    nfd = os.open("/dev/null", os.O_RDONLY)
    if params.get('quick', False) or re.match(r'[^/]+://', infile):
        retval = subprocess.call(ffmpeg_cmd + ['-i', infile] + ffmpeg_args + ffmpeg_args_url, close_fds=True, stderr=fp.fileno(), stdout=fp.fileno(), stdin=nfd)
    else:
        retval = subprocess.call(ffmpeg_cmd + ['-i', infile] + ffmpeg_args + ffmpeg_args_file, close_fds=True, stderr=fp.fileno(), stdout=fp.fileno(), stdin=nfd)
    os.close(nfd)
    fp.flush()
    fp.seek(0)
    bufl = fp.readlines()
    fp.close()
    is_input = False

    meta = {}

    for l0 in bufl:
      for l in l0.split(b"\r"):
        l = utils.toString(l)
        if re.search(r'^Input ', l):
            is_input = True
        if re.search(r'^Output ', l):
            is_input = False

        if is_input:
            m = re.search(r'Duration: ([0-9]+):([0-9]+):([0-9]+)\.([0-9]+)', l)
            if m:
                meta['meta_duration'] = int(m.group(1))*3600 + int(m.group(2))*60 + int(m.group(3)) + float("0." + m.group(4))

            m = re.search(r'Duration: .*, start: ([0-9.]+)', l)
            if m:
                meta['start'] = float(m.group(1))

            m = re.search(r'Stream.*Video:.* ([0-9]+)x([0-9]+)', l)
            if m:
                meta['width'] = int(m.group(1))
                meta['height'] = int(m.group(2))

            m = re.search(r'Stream.*Video:.* bitrate: ([^,]+)', l)
            if m:
                meta['video_bitrate'] = m.group(1)

            m = re.search(r'Stream.*Video: ([^,]+)', l)
            if m:
                meta['video_codec'] = m.group(1)

            m = re.search(r'Stream.*Video:.* ([0-9]+) tbc', l)
            if m:
                meta['frame_rate'] = m.group(1)

            m = re.search(r'Stream.*Audio:.* ([0-9]+) channels', l)
            if m:
                meta['audio_channels'] = int(m.group(1))

            m = re.search(r'Stream.*Audio:.* stereo', l)
            if m:
                meta['audio_channels'] = 2

            m = re.search(r'Stream.*Audio:.* mono', l)
            if m:
                meta['audio_channels'] = 1

            m = re.search(r'Stream.*Audio:.* Hz, ([0-9]+)\.([0-9]+),', l)
            if m:
                meta['audio_channels'] = int(m.group(1)) + int(m.group(2))

            m = re.search(r'Stream.*Audio:.* bitrate: ([^,]+)', l)
            if m:
                meta['audio_bitrate'] = m.group(1)

            m = re.search(r'Stream.*Audio: ([^,]+)', l)
            if m:
                meta['audio_codec'] = m.group(1)

            m = re.search(r'rotate *: ([0-9]+)', l)
            if m:
                meta['rotate'] = m.group(1)

        m = re.search(r' time=([0-9]+):([0-9]+):([0-9]+)\.([0-9]+) ', l)
        if m:
            meta['duration'] = int(m.group(1))*3600 + int(m.group(2))*60 + int(m.group(3)) + float("0." + m.group(4))

        m = re.search(r' time=([0-9]+)\.([0-9]+) ', l)
        if m:
            meta['duration'] = int(m.group(1)) + float("0." + m.group(2))


    if 'meta_duration' in meta and not 'duration' in meta:
        meta['duration'] = meta['meta_duration']

    return meta



import re

__all__ = ['plugin_task_lists']


TASK_LIST_ITEM = re.compile(r'^(\[[ xX]\])\s+')


def task_lists_hook(md, tokens, state):
    return _rewrite_all_list_items(tokens)


def render_ast_task_list_item(children, level, checked):
    return {
        'type': 'task_list_item',
        'children': children,
        'level': level,
        'checked': checked,
    }


def render_html_task_list_item(text, level, checked):
    checkbox = (
        '<input class="task-list-item-checkbox" '
        'type="checkbox"'
    )
    if checked:
        checkbox += ' checked/>'
    else:
        checkbox += '/>'

    if text.startswith('<p>'):
        text = text.replace('<p>', '<p>' + checkbox, 1)
    else:
        text = checkbox + text

    return '<li class="task-list-item">' + text + '</li>\n'


def plugin_task_lists(md):
    md.before_render_hooks.append(task_lists_hook)

    if md.renderer.NAME == 'html':
        md.renderer.register('task_list_item', render_html_task_list_item)
    elif md.renderer.NAME == 'ast':
        md.renderer.register('task_list_item', render_ast_task_list_item)


def _rewrite_all_list_items(tokens):
    for tok in tokens:
        if tok['type'] == 'list_item':
            _rewrite_list_item(tok)
        if 'children' in tok.keys():
            _rewrite_all_list_items(tok['children'])
    return tokens


def _rewrite_list_item(item):
    children = item['children']
    if children:
        first_child = children[0]
        text = first_child.get('text', '')
        m = TASK_LIST_ITEM.match(text)
        if m:
            mark = m.group(1)
            first_child['text'] = text[m.end():]

            params = item['params']
            if mark == '[ ]':
                params = (params[0], False)
            else:
                params = (params[0], True)

            item['type'] = 'task_list_item'
            item['params'] = params


from django.contrib import admin
from .models import *

admin.site.register(MediaLibrary)

admin.site.register(MediaItem)

admin.site.register(MediaItemRendition)

admin.site.register(MediaMeta)

class RenditionPresetAdmin(admin.ModelAdmin):
    list_display = ('name', 'short_name', 'stage')
    search_fields = ('name','sort_name')
admin.site.register(RenditionPreset, RenditionPresetAdmin)

class RenditionDefinitionAdmin(admin.ModelAdmin):
    list_display = ('name', 'preset', 'use')
    search_fields = ('name','use')
admin.site.register(RenditionDefinition, RenditionDefinitionAdmin)

admin.site.register(RenditionSet)

class RenditionParameterAdmin(admin.ModelAdmin):
    list_display = ('name', 'kind', 'required')
    search_fields = ('name',)
admin.site.register(RenditionParameter, RenditionParameterAdmin)

class RenditionParameterSettingAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'renditionPreset', 'parameter', 'setting')
    ordering = ('renditionPreset', 'parameter')
admin.site.register(RenditionPresetParameterSetting, RenditionParameterSettingAdmin)

class RenditionDefinitionSettingAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'renditionDefinition', 'parameter', 'setting')
    ordering = ('renditionDefinition', 'parameter')
admin.site.register(RenditionDefinitionParameterSetting, RenditionDefinitionSettingAdmin)

class RenderInstanceAdmin(admin.ModelAdmin):
    list_display = ('instance_id', 'priority', 'state', 'started', 'last_checkin', 'shutdown')
    ordering = ('shutdown', 'priority')
admin.site.register(RenderInstance, RenderInstanceAdmin)

class AccountConfigAdmin(admin.ModelAdmin):
    list_display = ('name', 'owner', 'applicable_preset')
    search_fields = ('name','owner')
admin.site.register(AccountConfig, AccountConfigAdmin)

class AccountConfigSettingAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'accountConfig', 'parameter', 'setting')
    ordering = ('accountConfig', 'parameter')
admin.site.register(AccountConfigParameterSetting, AccountConfigSettingAdmin)

class RenditionSegmentAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'rendition', 'start', 'end', 'duration')
    ordering = ('rendition', 'start')
admin.site.register(RenditionSegment, RenditionSegmentAdmin)



from .models import *

from sessionlog.models import SessionLog

from django.db.models.signals import post_init, m2m_changed
from django.db.models.fields.related import ManyToManyField
from django.core.exceptions import FieldDoesNotExist

from rest.middleware import get_request

import os
import sys
import inspect

try:
    from __main__ import AUDIT_DISABLED
except ImportError:
    AUDIT_DISABLED = False

def auditclass(fields=None, exclude=[]):
    """
    Audits field changes
    """

    def _store(self):
        if self.pk:
            self.__audit_old_data = dict((f, getattr(self, f)) for f in self.__audit_fields)
        else:
            self.__audit_old_data = {}

    def _post_init(sender, instance, **kwargs):
        _store(instance)

    def _logit(cls, pk, changes, skipstack=2):
        request = get_request()
        session = None
        user = None
        how = None
        referer = None

        if request:
            if request.user.is_authenticated:
                user = request.user
            session = SessionLog.GetSession(request)
            try:
                how = request.path
                referer = getattr(request, 'META', {}).get('HTTP_REFERER')
            except AttributeError:
                pass
            if not how:
                try:
                    how = sys.argv[0]
                    referer = ' '.join(sys.argv)
                except (AttributeError, IndexError):
                    pass
        
        stack = ', '.join(list("%s:%d:%s" % (os.path.basename(s[1]), s[2], s[3]) for s in inspect.stack()[skipstack:]))

        for k, v in list(changes.items()):
            AuditLog(
                model="%s.%s" % (cls._meta.app_label, cls._meta.model_name),
                pkey=pk,
                attribute=k,
                user=user,
                session=session,
                how=how,
                referer=referer,
                stack = stack,
                oldval = None if v[0] == None else str(v[0]),
                newval = None if v[1] == None else str(v[1]),
            ).save()

    def _m2m_changed(sender, instance, action, reverse, model, pk_set, using, **kwargs):
        oldval = None
        if action == 'post_clear':
            newval = 'CLEAR'
        elif action == 'post_add':
            newval = "ADD: "
            if not pk_set:
                return
        elif action == 'post_remove':
            newval = "REMOVE: "
            if not pk_set:
                return
        elif action == "pre_clear":
            pass
        else:
            return

        if reverse:
            mymodel = model
        else:
            mymodel = instance
        attname = None
        for f in mymodel._meta.many_to_many:
            if f.rel.through == sender:
                attname = f.attname
                break
        if not attname:
            return

        if action == "pre_clear":
            setattr(instance, '__audit_m2m_%s' % sender.__name__, ",".join(list(str(x) for x in getattr(instance, attname).all().values_list('pk', flat=True))))
            return
        if newval == "CLEAR":
            oldval = getattr(instance, '__audit_m2m_%s' % sender.__name__, '')

        if not pk_set:
            _logit(mymodel, instance.pk, {attname: (oldval, newval)}, skipstack=5)
        else:
            for pk in pk_set:
                if reverse:
                    _logit(mymodel, pk, {attname: (oldval, "%s%d" % (newval, instance.pk))}, skipstack=5)
                else:
                    _logit(mymodel, instance.pk, {attname: (oldval, "%s%d" % (newval, pk))}, skipstack=5)

    def inner(cls):
        # contains a local copy of the previous values of attributes
        cls.__audit_old_data = {}

        def has_changed(self, field):
            "Returns ``True`` if ``field`` has changed since initialization."
            return self.__audit_old_data.get(field) != getattr(self, field)
        cls.audit_has_changed = has_changed

        def old_value(self, field):
            "Returns the previous value of ``field``"
            return self.__audit_old_data.get(field)
        cls.audit_old_value = old_value

        def list_changed(self):
            "Returns a list of changed attributes."
            changed = {}
            for f in self.__audit_fields:
                if self.__audit_old_data.get(f) != getattr(self, f):
                    changed[f] = (self.__audit_old_data.get(f), getattr(self, f))
            return changed
        cls.audit_list_changed = list_changed

        # Ensure we are updating local attributes on model init
        post_init.connect(_post_init, sender=cls, weak=False)


        # Ensure we are updating local attributes on model save
        def save(self, *args, **kwargs):
            ret = save._original(self, *args, **kwargs)

            changes = self.audit_list_changed()
            if not changes:
                return ret

            _logit(self, self.pk, changes)
            _store(self)
            return ret
        save._original = cls.save
        cls.save = save

        def delete(self, *args, **kwargs):
            pk = self.pk
            ret = delete._original(self, *args, **kwargs)
            _logit(self, pk, {'*DELETE': (None,None,)})
            return ret
        delete._original = cls.delete
        cls.delete = delete
        
        if fields:
            afields = fields.copy()
        else:
            afields = []
            for f in cls._meta.fields + cls._meta.many_to_many:
                if getattr(f, 'auto_now', False) or getattr(f, 'auto_now_add', False) or getattr(f, 'primary_key', False):
                    continue
                afields.append(f.attname)
        for f in exclude:
            afields.remove(f)

        cls.__audit_fields = []
        cls.__audit_m2m_fields = []
        for f in afields:
            try:
                gf = cls._meta.get_field(f)
            except FieldDoesNotExist as e:
                if f[-3:] == '_id':
                    try:
                        gf = cls._meta.get_field(f[:-3])
                    except:
                        raise(e)
                else:
                    raise(e)
            if isinstance(gf, ManyToManyField):
                m2m_changed.connect(_m2m_changed, sender=gf.rel.through, weak=False)
                cls.__audit_m2m_fields.append(f)
            else:
                cls.__audit_fields.append(f)

            
        return cls

    if AUDIT_DISABLED:
        return lambda x: x
    return inner




# GEOPLUGIN API
_URL = "http://www.geoplugin.net/json.gp"
_URL_LAT = "http://www.geoplugin.net/extras/location.gp?lat={0}&long={1}&format=json"

def normalizeResponse(response):
    ret = {}
    data = json.loads(response)
    if "geoplugin_latitude" in data:
        ret["lat"] = data["geoplugin_latitude"]
        ret["lng"] = data["geoplugin_longitude"]
    if "geoplugin_city" in data:
        ret["city"] = data["geoplugin_city"]
    if "geoplugin_countryCode" in data:
        ret["country"] = data["geoplugin_countryCode"]
    if "geoplugin_region" in data:
        ret["region"] = data["geoplugin_region"]
    return ret

def _makeRequest(query_args, url):
    # This urlencodes your data (that's why we need to import urllib at the top)
    data = urllib.parse.urlencode(query_args)
    # Send HTTP POST request
    request = urllib.request.Request("{0}?{1}".format(url, data))
    try:
        return normalizeResponse(urllib.request.urlopen(request).read())
    except:
        pass
    return None

def byIP(ip):
    """
    simple geolocation by ip
      "geoplugin_request":"72.211.193.165",
      "geoplugin_status":200,
      "geoplugin_credit":"Some of the returned data includes GeoLite data created by MaxMind, available from <a href=\\'http:\/\/www.maxmind.com\\'>http:\/\/www.maxmind.com<\/a>.",
      "geoplugin_city":"San Clemente",
      "geoplugin_region":"CA",
      "geoplugin_areaCode":"949",
      "geoplugin_dmaCode":"803",
      "geoplugin_countryCode":"US",
      "geoplugin_countryName":"United States",
      "geoplugin_continentCode":"NA",
      "geoplugin_latitude":"33.422798",
      "geoplugin_longitude":"-117.620102",
      "geoplugin_regionCode":"CA",
      "geoplugin_regionName":"California",
      "geoplugin_currencyCode":"USD",
      "geoplugin_currencySymbol":"&#36;",
      "geoplugin_currencySymbol_UTF8":"$",
      "geoplugin_currencyConverter":1
    """
    # Prepare the data
    query_args = {
        'ip':ip,
    }

    # t = threading.Thread(target=_makeRequest, args=[query_args])
    # t.setDaemon(True)
    # t.start()
    return _makeRequest(query_args, _URL)



from rest import settings


DB_READWRITE_APPS = settings.get("DB_READWRITE_APPS", [])
DB_ROUTING_READ = settings.get("DB_ROUTING_READ", {})
DB_ROUTING_WRITE = settings.get("DB_ROUTING_WRITE", {})
DB_ROUTING_RELATIONS = settings.get("DB_ROUTING_RELATIONS", [('default', 'readonly')])
DB_ROUTING_MIGRATIONS = settings.get("DB_ROUTING_MIGRATIONS", {})
DB_ROUTING_MAPS = settings.get("DB_ROUTING_MAPS", {})
DB_ROUTING_READ_DEFAULT = settings.get("DB_ROUTING_READ_DEFAULT", "readonly")


class SimpleRouter:
    """
    A router that can be easily configured to route what database particular apps utilize
    """
    def getMappedDB(self, name):
        if DB_ROUTING_MAPS is not None and name in DB_ROUTING_MAPS:
            return DB_ROUTING_MAPS[name]
        return name

    def db_for_read(self, model, **hints):
        if "instance" in hints:
            instance = hints.get("instance", None)
            if instance and instance._state.db:
                return self.getMappedDB(instance._state.db)
        # check if this requires read/write dbs
        if model._meta.app_label in DB_READWRITE_APPS:
            return None
        # check if this is priority route
        if hasattr(model, "RestMeta") and getattr(model.RestMeta, "DB_PRIORITY", None):
            return None  
        # check if we have specific route
        for db_name, apps in DB_ROUTING_READ.items():
            if model._meta.app_label in apps:
                return self.getMappedDB(db_name)
        # default to readonly
        return self.getMappedDB(DB_ROUTING_READ_DEFAULT)

    def db_for_write(self, model, **hints):
        # check if we have specific route
        for db_name, apps in DB_ROUTING_WRITE.items():
            if model._meta.app_label in apps:
                return self.getMappedDB(db_name)
        return "default"

    def allow_relation(self, obj1, obj2, **hints):
        for rel in DB_ROUTING_RELATIONS:
            if obj1._state.db in rel and obj2._state.db in rel:
                return True
        return False

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        apps = DB_ROUTING_MIGRATIONS.get(db, None)
        if apps is None:
            return None  # no preference
        if app_label in apps:
            return True  # allow migrations
        return False  # do not allow migrations



import requests
from rest import settings
from rest import log

# SEND_URL = "POST https://fcm.googleapis.com/v1/{parent=projects/*}/messages:send"
SEND_URL = "https://fcm.googleapis.com/fcm/send"
CM_PROVIDERS = settings.get("CM_PROVIDERS", None)

FCM_KEY = CM_PROVIDERS["fcm"]["key"]

logger = log.getLogger("fcm", filename="fcm.log")


def sendToDevice(device, data):
    return sendData(device.cm_token, data)


def sendNotification(to_token, title, body):
    return postMessage(dict(to=to_token, notification=dict(title=title, body=body)))


def sendData(to_token, data, priority="high"):
    return postMessage(dict(to=to_token, data=data, content_available=True, priority=priority))


def postMessage(payload):
    logger.info("sending FCM", payload)
    headers = dict(Authorization=F"key={FCM_KEY}")
    headers['Content-type'] = 'application/json'
    resp = requests.post(SEND_URL, json=payload, headers=headers)
    logger.info("response", resp.text)
    return resp




from django.db import models
from rest import settings

from rest.models import RestModel
from rest import mail as rest_mail
from rest import helpers as rh
import inbox.utils

from .member import Member

from datetime import datetime, timedelta
import threading


def sendEmail(email_to, subject, body=None, attachments=[], 
              from_email=settings.DEFAULT_FROM_EMAIL, background=False):
    record = recordEmail(email_to, subject, body, attachments, from_email)
    record.send(background=background)
    return record


def recordEmail(email_to, subject, body=None, attachments=[], 
                from_email=settings.DEFAULT_FROM_EMAIL):
    email_record = NotificationRecord(
        method="email",
        subject=subject,
        from_addr=from_email,
        body=body)
    email_record.save()
    email_record.addAttachments(attachments)
    if isinstance(email_to, str):
        email_to = [email_to]
    for email in email_to:
        member = Member.objects.filter(email=email).last()
        nr = NotificationMemberRecord(
            member=member, to_addr=email,
            notification=email_record)
        nr.save()
    return email_record


class NotificationRecord(models.Model, RestModel):
    class RestMeta:
        CAN_SAVE = CAN_CREATE = False
        DEFAULT_SORT = "-created"
        SEARCH_FIELDS = ["subject"]
        SEARCH_TERMS = ["subject", ("to", "to__to_addr"), "body", "reason", "state", ("from", "from_addr")]
        GRAPHS = {
            "list": {
                "fields": ["id", ("get_state_display", "state_display"), "created", "subject", "from_addr", "to_emails", "reason", "state", "attempts"],
            },
            "default": {
                "extra": ["to_emails", ("get_state_display", "state_display")]
            }
        }
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    modified = models.DateTimeField(auto_now=True, editable=False)
    method = models.CharField(max_length=128, default="email", db_index=True)

    from_addr = models.CharField(max_length=255)
    subject = models.CharField(max_length=255)
    body = models.TextField()
    reason = models.TextField()
    # new=0, queued=-5, sent=1
    state = models.IntegerField(default=0, choices=[(0, "new"), (-5, "queued"), (1, "sent"), (-10, "failed")], db_index=True)
    attempts = models.IntegerField(default=0)

    @property
    def to_emails(self):
        return list(self.to.all().values_list("to_addr", flat=True))

    def send(self, member_records=None, background=True):
        email_to = []
        save_records = True
        if not member_records:
            member_records = self.to.all()
            save_records = False
        for r in member_records:
            email_to.append(r.to_addr)
            if save_records:
                r.notification = self
                r.save()
        if NotificationRecord.canSend():
            if background:
                t = threading.Thread(target=self.sendNow, args=[email_to])
                t.start()
            else:
                self.sendNow(email_to)
            return True
        if self.state != -5:
            self.state = -5
            self.save()
        return False

    def sendNow(self, email_to):
        try:
            inbox.utils.send(
                self.from_addr,
                email_to,
                self.subject,
                self.body,
                attachments=self.attachments.all(),
                fail_silently=False)
            # self.reason = "sent"
            self.state = 1
        except Exception as err:
            self.reason = str(err)
            self.attempts += 1
            if self.attempts >= 3:
                self.state = -10
            else:
                self.state = -5
        self.save()
        return True 

    def attach(self, name, mimetype, data):
        atmnt = NotificationAttachment(notification=self, name=name, mimetype=mimetype, data=data)
        atmnt.save()
        return atmnt

    def addAttachments(self, attachments):
        if not attachments:
            return False
        for a in attachments:
            if isinstance(a, str):
                # TODO handle file inport
                pass
            else:
                self.attach(a.name, a.mimetype, a.data)

    @classmethod
    def canSend(cls):
        max_emails_per_minute = settings.get("MAX_EMAILS_PER_MINUTE", 30)
        last_email = NotificationRecord.objects.filter(state=1).last()
        now = datetime.now()
        if last_email and (now - last_email.created).total_seconds() < 30:
            # we sent an email less then a minute ago
            # now we can to count the number of message sent in last minute
            when = now - timedelta(seconds=60)
            sent = NotificationRecord.objects.filter(state=1, created__gte=when).count()
            return sent < max_emails_per_minute
        return True

    @classmethod
    def notifyFromEmails(cls, emails, subject, message=None,
                         template=None, context=None, email_only=False,
                         sms_msg=None, force=False,
                         from_email=settings.DEFAULT_FROM_EMAIL,
                         attachments=[]):
        # Member = RestModel.getModel("account", "Member")
        members = Member.objects.filter(email__in=emails)
        cls.notify(members, subject, message, template, context, email_only, sms_msg, force, from_email, attachments)

    @classmethod
    def notify(cls, notify_users, subject, message=None, 
               template=None, context=None, email_only=False,
               sms_msg=None, force=False,
               from_email=settings.DEFAULT_FROM_EMAIL,
               attachments=[]):
        # this will create a record for each email address message is sent to
        from telephony.models import SMS
        email_to = []
        email_list = []
        sms_to = []

        if not message and not template and subject:
            message = subject

        if not sms_msg and subject:
            sms_msg = subject
        if not sms_msg and message:
            sms_msg = message

        if subject and len(subject) > 80:
            epos = subject.find('. ') + 1
            if epos > 10:
                subject = subject[:epos]
            else:
                subject = subject[:80]
                subject = subject[:subject.rfind(' ')] + "..."

        if template:
            # render message now so we can save message
            if context is None:
                context = {}
            if message is not None:
                context["body"] = message
            message = inbox.utils.renderTemplate(template, context)
            # message = rest_mail.renderBody(message, template, context)
            template = None
            context = None

        email_record = None
        for member in notify_users:
            via = member.getProperty("notify_via", "all")
            phone = member.getProperty("phone")
            email = member.email
            valid_email = email is not None and "@" in email and "invalid" not in email
            allow_sms = not email_only and phone and (force or via in ["all", "sms"])
            allow_email = valid_email and (force or via in ["all", "email"])
            if not allow_email and not allow_sms:
                continue
            if allow_email and email not in email_list:
                email_list.append(email)
                nr = NotificationMemberRecord(member=member, to_addr=email)
                email_to.append(nr)
            if not email_only and allow_sms and phone not in sms_to:
                sms_to.append(phone)

        if sms_to:
            for phone in sms_to:
                SMS.send(phone, sms_msg)

        if email_to:
            # lets verify the db is working
            email_record = NotificationRecord(
                method="email",
                subject=subject,
                from_addr=from_email,
                body=message)
            try:
                email_record.save()
                email_record.addAttachments(attachments)
                email_record.send(email_to)
            except Exception as err:
                print(("failed to create record: {}".format(str(err))))
                # we need to send emails the old way
                addrs = []
                for to in email_to:
                    addrs.append(to.to_addr)
                rest_mail.send(
                    addrs,
                    subject,
                    message,
                    attachments=attachments,
                    do_async=True)


class NotificationAttachment(models.Model, RestModel):
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    notification = models.ForeignKey(NotificationRecord, related_name="attachments", on_delete=models.CASCADE)
    name = models.CharField(max_length=255)
    mimetype = models.CharField(max_length=255)
    data = models.TextField()


class NotificationMemberRecord(models.Model, RestModel):
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    member = models.ForeignKey(
        "account.Member", related_name="notifications",
        default=None, null=True, on_delete=models.CASCADE)
    notification = models.ForeignKey(NotificationRecord, related_name="to", on_delete=models.CASCADE)
    to_addr = models.CharField(max_length=255, db_index=True)


class BounceHistory(models.Model, RestModel):
    class RestMeta:
        CAN_SAVE = False
        SEARCH_FIELDS = ["address"]
        SEARCH_TERMS = [("email", "address"), ("to", "address"), "source", "reason", "state", ("user", "user__username")]
        GRAPHS = {
            "default": {
                "graphs": {
                    "user": "basic"
                }
            },
            "list": {
                "graphs": {
                    "user": "basic"
                }
            }
        }
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    user = models.ForeignKey("account.Member", related_name="bounces", null=True, blank=True, default=None, on_delete=models.CASCADE)
    address = models.CharField(max_length=255, db_index=True)
    kind = models.CharField(max_length=32, db_index=True)
    reason = models.TextField(null=True, blank=True, default=None)
    reporter = models.CharField(max_length=255, null=True, blank=True, default=None)
    code = models.CharField(max_length=32, null=True, blank=True, default=None)
    source = models.CharField(max_length=255, null=True, blank=True, default=None)
    source_ip = models.CharField(max_length=64, null=True, blank=True, default=None)

    @staticmethod
    def log(kind, address, reason, reporter=None, code=None, source=None, source_ip=None, user=None):
        obj = BounceHistory(kind=kind, address=address)
        obj.reason = reason
        obj.reporter = reporter
        obj.code = code
        obj.source = source
        obj.source_ip = source_ip
        if user is None:
            Member = RestModel.getModel("account", "Member")
            user = Member.objects.filter(email=address).last()
            # now lets check our bounced count, if more then 3, we turn off email
            if user:
                user.log("bounced", "{} bounced to {} from {}".format(kind, address, source_ip), method=kind)
                since = datetime.now() - timedelta(days=14)
                bounce_count = BounceHistory.objects.filter(user=user, created__gte=since).count()
                if bounce_count > 2:
                    # TODO notify support an account has been disabled because of bounce
                    user.setProperty("notify_via", "off")
                    user.log("disabled", "notifications disabled because email bounced", method="notify")
        else:
            # TODO notify support of unknown bounce
            pass
        obj.user = user
        obj.save()




# Generated by Django 4.1.3 on 2022-12-15 00:13

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0012_settings_settingsmetadata'),
        ('medialib', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Mailbox',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('email', models.CharField(db_index=True, max_length=255)),
                ('state', models.IntegerField(db_index=True, default=0)),
                ('tq_app', models.CharField(default=None, max_length=255, null=True)),
                ('tq_handler', models.CharField(default=None, max_length=255, null=True)),
                ('tq_channel', models.CharField(default=None, max_length=255, null=True)),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('sent_at', models.DateTimeField()),
                ('state', models.IntegerField(db_index=True, default=0)),
                ('to_email', models.CharField(db_index=True, max_length=255)),
                ('to_name', models.CharField(default=None, max_length=255, null=True)),
                ('to', models.TextField()),
                ('cc', models.TextField()),
                ('from_email', models.CharField(db_index=True, max_length=255)),
                ('from_name', models.CharField(default=None, max_length=255, null=True)),
                ('subject', models.CharField(default=None, max_length=255, null=True)),
                ('message', models.TextField(default=None, null=True)),
                ('body', models.TextField(default=None, null=True)),
                ('html', models.TextField(default=None, null=True)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='MailboxMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='inbox.mailbox')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Complaint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('address', models.CharField(db_index=True, max_length=255)),
                ('kind', models.CharField(db_index=True, max_length=32)),
                ('reason', models.TextField(blank=True, default=None, null=True)),
                ('user_agent', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('source', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('source_ip', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('user', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='emails_complaints', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='Bounce',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('address', models.CharField(db_index=True, max_length=255)),
                ('kind', models.CharField(db_index=True, max_length=32)),
                ('reason', models.TextField(blank=True, default=None, null=True)),
                ('reporter', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('code', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('source', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('source_ip', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('user', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='emails_bounced', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('name', models.CharField(default=None, max_length=255, null=True)),
                ('media', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='medialib.mediaitem')),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='inbox.message')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]



# coding: utf-8

"""
    mistune_contrib.highlight
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Support highlight code features for mistune.

    :copyright: (c) 2014 - 2015 by Hsiaoming Yang.
"""

import mistune
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter


def block_code(text, lang, inlinestyles=False, linenos=False):
    if not lang:
        text = text.strip()
        return '<pre><code>%s</code></pre>\n' % mistune.escape(text)

    try:
        lexer = get_lexer_by_name(lang, stripall=True)
        formatter = HtmlFormatter(
            noclasses=inlinestyles, linenos=linenos
        )
        code = highlight(text, lexer, formatter)
        if linenos:
            return '<div class="highlight-wrapper">%s</div>\n' % code
        return code
    except:
        return '<pre class="%s"><code>%s</code></pre>\n' % (
            lang, mistune.escape(text)
        )


class HighlightMixin(object):
    def block_code(self, text, language="text"):
        # renderer has an options
        inlinestyles = False  #  self.options.get('inlinestyles', False)
        linenos = False  #  self.options.get('linenos', False)
        return block_code(text, language, inlinestyles, linenos)



#!/usr/bin/python2.4
#
# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model objects for requests and responses.

Each API may support one or more serializations, such
as JSON, Atom, etc. The model classes are responsible
for converting between the wire format and the Python
object representation.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import logging
import urllib.request, urllib.parse, urllib.error

from medialib.youtube.apiclient import __version__
from .errors import HttpError
from medialib.youtube.oauth2client.anyjson import simplejson


dump_request_response = False


def _abstract():
  raise NotImplementedError('You need to override this function')


class Model(object):
  """Model base class.

  All Model classes should implement this interface.
  The Model serializes and de-serializes between a wire
  format such as JSON and a Python object representation.
  """

  def request(self, headers, path_params, query_params, body_value):
    """Updates outgoing requests with a serialized body.

    Args:
      headers: dict, request headers
      path_params: dict, parameters that appear in the request path
      query_params: dict, parameters that appear in the query
      body_value: object, the request body as a Python object, which must be
                  serializable.
    Returns:
      A tuple of (headers, path_params, query, body)

      headers: dict, request headers
      path_params: dict, parameters that appear in the request path
      query: string, query part of the request URI
      body: string, the body serialized in the desired wire format.
    """
    _abstract()

  def response(self, resp, content):
    """Convert the response wire format into a Python object.

    Args:
      resp: httplib2.Response, the HTTP response headers and status
      content: string, the body of the HTTP response

    Returns:
      The body de-serialized as a Python object.

    Raises:
      apiclient.errors.HttpError if a non 2xx response is received.
    """
    _abstract()


class BaseModel(Model):
  """Base model class.

  Subclasses should provide implementations for the "serialize" and
  "deserialize" methods, as well as values for the following class attributes.

  Attributes:
    accept: The value to use for the HTTP Accept header.
    content_type: The value to use for the HTTP Content-type header.
    no_content_response: The value to return when deserializing a 204 "No
        Content" response.
    alt_param: The value to supply as the "alt" query parameter for requests.
  """

  accept = None
  content_type = None
  no_content_response = None
  alt_param = None

  def _log_request(self, headers, path_params, query, body):
    """Logs debugging information about the request if requested."""
    if dump_request_response:
      logging.info('--request-start--')
      logging.info('-headers-start-')
      for h, v in list(headers.items()):
        logging.info('%s: %s', h, v)
      logging.info('-headers-end-')
      logging.info('-path-parameters-start-')
      for h, v in list(path_params.items()):
        logging.info('%s: %s', h, v)
      logging.info('-path-parameters-end-')
      logging.info('body: %s', body)
      logging.info('query: %s', query)
      logging.info('--request-end--')

  def request(self, headers, path_params, query_params, body_value):
    """Updates outgoing requests with a serialized body.

    Args:
      headers: dict, request headers
      path_params: dict, parameters that appear in the request path
      query_params: dict, parameters that appear in the query
      body_value: object, the request body as a Python object, which must be
                  serializable by simplejson.
    Returns:
      A tuple of (headers, path_params, query, body)

      headers: dict, request headers
      path_params: dict, parameters that appear in the request path
      query: string, query part of the request URI
      body: string, the body serialized as JSON
    """
    query = self._build_query(query_params)
    headers['accept'] = self.accept
    headers['accept-encoding'] = 'gzip, deflate'
    if 'user-agent' in headers:
      headers['user-agent'] += ' '
    else:
      headers['user-agent'] = ''
    headers['user-agent'] += 'google-api-python-client/%s (gzip)' % __version__

    if body_value is not None:
      headers['content-type'] = self.content_type
      body_value = self.serialize(body_value)
    self._log_request(headers, path_params, query, body_value)
    return (headers, path_params, query, body_value)

  def _build_query(self, params):
    """Builds a query string.

    Args:
      params: dict, the query parameters

    Returns:
      The query parameters properly encoded into an HTTP URI query string.
    """
    if self.alt_param is not None:
      params.update({'alt': self.alt_param})
    astuples = []
    for key, value in list(params.items()):
      if type(value) == type([]):
        for x in value:
          x = x.encode('utf-8')
          astuples.append((key, x))
      else:
        if getattr(value, 'encode', False) and callable(value.encode):
          value = value.encode('utf-8')
        astuples.append((key, value))
    return '?' + urllib.parse.urlencode(astuples)

  def _log_response(self, resp, content):
    """Logs debugging information about the response if requested."""
    if dump_request_response:
      logging.info('--response-start--')
      for h, v in list(resp.items()):
        logging.info('%s: %s', h, v)
      if content:
        logging.info(content)
      logging.info('--response-end--')

  def response(self, resp, content):
    """Convert the response wire format into a Python object.

    Args:
      resp: httplib2.Response, the HTTP response headers and status
      content: string, the body of the HTTP response

    Returns:
      The body de-serialized as a Python object.

    Raises:
      apiclient.errors.HttpError if a non 2xx response is received.
    """
    self._log_response(resp, content)
    # Error handling is TBD, for example, do we retry
    # for some operation/error combinations?
    if resp.status < 300:
      if resp.status == 204:
        # A 204: No Content response should be treated differently
        # to all the other success states
        return self.no_content_response
      return self.deserialize(content)
    else:
      logging.debug('Content from bad request was: %s' % content)
      raise HttpError(resp, content)

  def serialize(self, body_value):
    """Perform the actual Python object serialization.

    Args:
      body_value: object, the request body as a Python object.

    Returns:
      string, the body in serialized form.
    """
    _abstract()

  def deserialize(self, content):
    """Perform the actual deserialization from response string to Python
    object.

    Args:
      content: string, the body of the HTTP response

    Returns:
      The body de-serialized as a Python object.
    """
    _abstract()


class JsonModel(BaseModel):
  """Model class for JSON.

  Serializes and de-serializes between JSON and the Python
  object representation of HTTP request and response bodies.
  """
  accept = 'application/json'
  content_type = 'application/json'
  alt_param = 'json'

  def __init__(self, data_wrapper=False):
    """Construct a JsonModel.

    Args:
      data_wrapper: boolean, wrap requests and responses in a data wrapper
    """
    self._data_wrapper = data_wrapper

  def serialize(self, body_value):
    if (isinstance(body_value, dict) and 'data' not in body_value and
        self._data_wrapper):
      body_value = {'data': body_value}
    return simplejson.dumps(body_value)

  def deserialize(self, content):
    content = content.decode('utf-8')
    body = simplejson.loads(content)
    if self._data_wrapper and isinstance(body, dict) and 'data' in body:
      body = body['data']
    return body

  @property
  def no_content_response(self):
    return {}


class RawModel(JsonModel):
  """Model class for requests that don't return JSON.

  Serializes and de-serializes between JSON and the Python
  object representation of HTTP request, and returns the raw bytes
  of the response body.
  """
  accept = '*/*'
  content_type = 'application/json'
  alt_param = None

  def deserialize(self, content):
    return content

  @property
  def no_content_response(self):
    return ''


class MediaModel(JsonModel):
  """Model class for requests that return Media.

  Serializes and de-serializes between JSON and the Python
  object representation of HTTP request, and returns the raw bytes
  of the response body.
  """
  accept = '*/*'
  content_type = 'application/json'
  alt_param = 'media'

  def deserialize(self, content):
    return content

  @property
  def no_content_response(self):
    return ''


class ProtocolBufferModel(BaseModel):
  """Model class for protocol buffers.

  Serializes and de-serializes the binary protocol buffer sent in the HTTP
  request and response bodies.
  """
  accept = 'application/x-protobuf'
  content_type = 'application/x-protobuf'
  alt_param = 'proto'

  def __init__(self, protocol_buffer):
    """Constructs a ProtocolBufferModel.

    The serialzed protocol buffer returned in an HTTP response will be
    de-serialized using the given protocol buffer class.

    Args:
      protocol_buffer: The protocol buffer class used to de-serialize a
      response from the API.
    """
    self._protocol_buffer = protocol_buffer

  def serialize(self, body_value):
    return body_value.SerializeToString()

  def deserialize(self, content):
    return self._protocol_buffer.FromString(content)

  @property
  def no_content_response(self):
    return self._protocol_buffer()


def makepatch(original, modified):
  """Create a patch object.

  Some methods support PATCH, an efficient way to send updates to a resource.
  This method allows the easy construction of patch bodies by looking at the
  differences between a resource before and after it was modified.

  Args:
    original: object, the original deserialized resource
    modified: object, the modified deserialized resource
  Returns:
    An object that contains only the changes from original to modified, in a
    form suitable to pass to a PATCH method.

  Example usage:
    item = service.activities().get(postid=postid, userid=userid).execute()
    original = copy.deepcopy(item)
    item['object']['content'] = 'This is updated.'
    service.activities.patch(postid=postid, userid=userid,
      body=makepatch(original, item)).execute()
  """
  patch = {}
  for key, original_value in list(original.items()):
    modified_value = modified.get(key, None)
    if modified_value is None:
      # Use None to signal that the element is deleted
      patch[key] = None
    elif original_value != modified_value:
      if type(original_value) == type({}):
        # Recursively descend objects
        patch[key] = makepatch(original_value, modified_value)
      else:
        # In the case of simple types or arrays we just replace
        patch[key] = modified_value
    else:
      # Don't add anything to patch if there's no change
      pass
  for key in modified:
    if key not in original:
      patch[key] = modified[key]

  return patch



from .member import *  # noqa: F401, F403
from .group import *  # noqa: F401, F403
from .membership import *  # noqa: F401, F403
from .notify import *  # noqa: F401, F403
from .feeds import *  # noqa: F401, F403
from .session import *  # noqa: F401, F403
from .device import *  # noqa: F401, F403
from .settings import *  # noqa: F401, F403



# Generated by Django 4.1.4 on 2023-06-23 22:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('metrics', '0004_eodmetrics'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metrics',
            name='v1',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v10',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v11',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v12',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v13',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v14',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v2',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v3',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v4',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v5',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v6',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v7',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v8',
            field=models.BigIntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='metrics',
            name='v9',
            field=models.BigIntegerField(default=0),
        ),
    ]



"""
This command essentially migrates data used for a number of items used by
the R class. This should fix any issues related to:

    WRONGTYPE Operation against a key holding the wrong kind of value

These include:

*

"""
from __future__ import unicode_literals
import json

from django.core.management.base import NoArgsCommand
from redis.exceptions import ResponseError
from redis_metrics.utils import get_r


class Command(NoArgsCommand):
    help = "Fixes backward-incompatible changes to the way we store some keys in Redis"

    def handle_noargs(self, *args, **options):
        r = get_r()

        # Store *only* slugs in the Set of metric slugs. Prior to this, we
        # stored metric keys, and split the slug from the key upon retrieval
        try:
            keys = r.r.smembers(r._metric_slugs_key)
            slugs = set(s.split(":")[1] for s in keys)
            r.r.srem(r._metric_slugs_key, *keys)  # Remove the old keys
            r.r.sadd(r._metric_slugs_key, *slugs)  # Keep the new slugs

            p = "\nMetrics: Converted {0} Keys to {1} Slugs\n"
            self.stdout.write(p.format(len(keys), len(slugs)))
        except (IndexError, ResponseError):
            # If none of our old keys contain a ':', we don't need to replace them.
            pass

        # Similar to metric keys above...
        # Store *only* slugs in the Set of Gauge slugs. Prior to this, we
        # stored gauge keys, and split the slug from the key upon retrieval
        try:
            keys = r.r.smembers(r._gauge_slugs_key)
            slugs = set(s.split(":")[1] for s in keys)
            r.r.srem(r._gauge_slugs_key, *keys)  # Remove the old keys
            r.r.sadd(r._gauge_slugs_key, *slugs)  # Keep the new slugs

            p = "Gauges: Converted {0} Keys to {1} Slugs\n"
            self.stdout.write(p.format(len(keys), len(slugs)))
        except (IndexError, ResponseError):
            # If none of our old keys contain a ':', we don't need to replace them.
            pass

        # Convert all Categories from JSON to Sets.
        i = 0
        categories = r.categories()
        for category in categories:
            try:
                k = r._category_key(category)
                data = r.r.get(k)  # Get the JSON-ecoded list of metrics in this Category
                if data:  # redis gives us a None-value for a non-existing key
                    data = json.loads(data)  # convert to python
                    r.r.delete(k)  # Delete the existing info in redis
                    r.r.sadd(k, *set(data))  # Re-create as a Set
                    i += 1
            except ResponseError:
                pass

        if i > 0:
            p = "Converted {0} Categories from JSON -> Redis Sets\n"
            self.stdout.write(p.format(i))




from datetime import datetime
from rest import mail

def SEND(task, attempts=3):
    now = datetime.now()
    filename = task.data.filename
    if filename and '{date' in filename:
        filename = filename.format(date=now)
    subject = task.data.subject
    if subject and '{date' in subject:
        subject = subject.format(date=now)
    atch1 = mail.makeAttachment(filename, task.data.data)
    try:
        mail.send(task.data.address, subject, attachments=[atch1], body=task.data.body, template=task.data.template)
    except Exception as err:
        task.log_exception(err)
        return False
    return True



from rest import decorators as rd
from .. import models as location


@rd.urlGET('ip')
def on_myip(request):
    gip = location.GeoIP.get(request.ip)
    graph = request.DATA.get("graph", "default")
    return gip.restGet(request, graph)


@rd.urlGET('ips')
def rest_on_geoip(request):
    return location.GeoIP.on_rest_list(request)


@rd.urlGET(r'^ip/lookup$')
def on_ip_lookup(request):
    ip = request.DATA.get("ip", request.ip)
    gip = location.GeoIP.get(
        ip,
        force_refresh=request.DATA.get("refresh", 0, field_type=bool),
        stale_after=request.DATA.get("stale_after", 90, field_type=int))
    graph = request.DATA.get("graph", "default")
    return gip.restGet(request, graph)


@rd.url(r'^geo/ip$')
@rd.url(r'^geo/ip/(?P<pk>\d+)$')
@rd.login_required
def rest_on_geoip(request, pk=None):
    return location.GeoIP.on_rest_request(request, pk)



#!/usr/bin/python2.4
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import hashlib
import logging
import time

from .anyjson import simplejson


CLOCK_SKEW_SECS = 300  # 5 minutes in seconds
AUTH_TOKEN_LIFETIME_SECS = 300  # 5 minutes in seconds
MAX_TOKEN_LIFETIME_SECS = 86400  # 1 day in seconds


logger = logging.getLogger(__name__)


class AppIdentityError(Exception):
  pass


try:
  from OpenSSL import crypto


  class OpenSSLVerifier(object):
    """Verifies the signature on a message."""

    def __init__(self, pubkey):
      """Constructor.

      Args:
        pubkey, OpenSSL.crypto.PKey, The public key to verify with.
      """
      self._pubkey = pubkey

    def verify(self, message, signature):
      """Verifies a message against a signature.

      Args:
        message: string, The message to verify.
        signature: string, The signature on the message.

      Returns:
        True if message was signed by the private key associated with the public
        key that this object was constructed with.
      """
      try:
        crypto.verify(self._pubkey, signature, message, 'sha256')
        return True
      except:
        return False

    @staticmethod
    def from_string(key_pem, is_x509_cert):
      """Construct a Verified instance from a string.

      Args:
        key_pem: string, public key in PEM format.
        is_x509_cert: bool, True if key_pem is an X509 cert, otherwise it is
          expected to be an RSA key in PEM format.

      Returns:
        Verifier instance.

      Raises:
        OpenSSL.crypto.Error if the key_pem can't be parsed.
      """
      if is_x509_cert:
        pubkey = crypto.load_certificate(crypto.FILETYPE_PEM, key_pem)
      else:
        pubkey = crypto.load_privatekey(crypto.FILETYPE_PEM, key_pem)
      return OpenSSLVerifier(pubkey)


  class OpenSSLSigner(object):
    """Signs messages with a private key."""

    def __init__(self, pkey):
      """Constructor.

      Args:
        pkey, OpenSSL.crypto.PKey (or equiv), The private key to sign with.
      """
      self._key = pkey

    def sign(self, message):
      """Signs a message.

      Args:
        message: string, Message to be signed.

      Returns:
        string, The signature of the message for the given key.
      """
      return crypto.sign(self._key, message, 'sha256')

    @staticmethod
    def from_string(key, password='notasecret'):
      """Construct a Signer instance from a string.

      Args:
        key: string, private key in PKCS12 or PEM format.
        password: string, password for the private key file.

      Returns:
        Signer instance.

      Raises:
        OpenSSL.crypto.Error if the key can't be parsed.
      """
      if key.startswith('-----BEGIN '):
        pkey = crypto.load_privatekey(crypto.FILETYPE_PEM, key)
      else:
        pkey = crypto.load_pkcs12(key, password).get_privatekey()
      return OpenSSLSigner(pkey)

except ImportError:
  OpenSSLVerifier = None
  OpenSSLSigner = None


try:
  from Crypto.PublicKey import RSA
  from Crypto.Hash import SHA256
  from Crypto.Signature import PKCS1_v1_5


  class PyCryptoVerifier(object):
    """Verifies the signature on a message."""

    def __init__(self, pubkey):
      """Constructor.

      Args:
        pubkey, OpenSSL.crypto.PKey (or equiv), The public key to verify with.
      """
      self._pubkey = pubkey

    def verify(self, message, signature):
      """Verifies a message against a signature.

      Args:
        message: string, The message to verify.
        signature: string, The signature on the message.

      Returns:
        True if message was signed by the private key associated with the public
        key that this object was constructed with.
      """
      try:
        return PKCS1_v1_5.new(self._pubkey).verify(
            SHA256.new(message), signature)
      except:
        return False

    @staticmethod
    def from_string(key_pem, is_x509_cert):
      """Construct a Verified instance from a string.

      Args:
        key_pem: string, public key in PEM format.
        is_x509_cert: bool, True if key_pem is an X509 cert, otherwise it is
          expected to be an RSA key in PEM format.

      Returns:
        Verifier instance.

      Raises:
        NotImplementedError if is_x509_cert is true.
      """
      if is_x509_cert:
        raise NotImplementedError(
            'X509 certs are not supported by the PyCrypto library. '
            'Try using PyOpenSSL if native code is an option.')
      else:
        pubkey = RSA.importKey(key_pem)
      return PyCryptoVerifier(pubkey)


  class PyCryptoSigner(object):
    """Signs messages with a private key."""

    def __init__(self, pkey):
      """Constructor.

      Args:
        pkey, OpenSSL.crypto.PKey (or equiv), The private key to sign with.
      """
      self._key = pkey

    def sign(self, message):
      """Signs a message.

      Args:
        message: string, Message to be signed.

      Returns:
        string, The signature of the message for the given key.
      """
      return PKCS1_v1_5.new(self._key).sign(SHA256.new(message))

    @staticmethod
    def from_string(key, password='notasecret'):
      """Construct a Signer instance from a string.

      Args:
        key: string, private key in PEM format.
        password: string, password for private key file. Unused for PEM files.

      Returns:
        Signer instance.

      Raises:
        NotImplementedError if they key isn't in PEM format.
      """
      if key.startswith('-----BEGIN '):
        pkey = RSA.importKey(key)
      else:
        raise NotImplementedError(
            'PKCS12 format is not supported by the PyCrpto library. '
            'Try converting to a "PEM" '
            '(openssl pkcs12 -in xxxxx.p12 -nodes -nocerts > privatekey.pem) '
            'or using PyOpenSSL if native code is an option.')
      return PyCryptoSigner(pkey)

except ImportError:
  PyCryptoVerifier = None
  PyCryptoSigner = None


if OpenSSLSigner:
  Signer = OpenSSLSigner
  Verifier = OpenSSLVerifier
elif PyCryptoSigner:
  Signer = PyCryptoSigner
  Verifier = PyCryptoVerifier
else:
  raise ImportError('No encryption library found. Please install either '
                    'PyOpenSSL, or PyCrypto 2.6 or later')


def _urlsafe_b64encode(raw_bytes):
  return base64.urlsafe_b64encode(raw_bytes).rstrip('=')


def _urlsafe_b64decode(b64string):
  # Guard against unicode strings, which base64 can't handle.
  b64string = b64string.encode('ascii')
  padded = b64string + '=' * (4 - len(b64string) % 4)
  return base64.urlsafe_b64decode(padded)


def _json_encode(data):
  return simplejson.dumps(data, separators = (',', ':'))


def make_signed_jwt(signer, payload):
  """Make a signed JWT.

  See http://self-issued.info/docs/draft-jones-json-web-token.html.

  Args:
    signer: crypt.Signer, Cryptographic signer.
    payload: dict, Dictionary of data to convert to JSON and then sign.

  Returns:
    string, The JWT for the payload.
  """
  header = {'typ': 'JWT', 'alg': 'RS256'}

  segments = [
          _urlsafe_b64encode(_json_encode(header)),
          _urlsafe_b64encode(_json_encode(payload)),
  ]
  signing_input = '.'.join(segments)

  signature = signer.sign(signing_input)
  segments.append(_urlsafe_b64encode(signature))

  logger.debug(str(segments))

  return '.'.join(segments)


def verify_signed_jwt_with_certs(jwt, certs, audience):
  """Verify a JWT against public certs.

  See http://self-issued.info/docs/draft-jones-json-web-token.html.

  Args:
    jwt: string, A JWT.
    certs: dict, Dictionary where values of public keys in PEM format.
    audience: string, The audience, 'aud', that this JWT should contain. If
      None then the JWT's 'aud' parameter is not verified.

  Returns:
    dict, The deserialized JSON payload in the JWT.

  Raises:
    AppIdentityError if any checks are failed.
  """
  segments = jwt.split('.')

  if (len(segments) != 3):
    raise AppIdentityError(
      'Wrong number of segments in token: %s' % jwt)
  signed = '%s.%s' % (segments[0], segments[1])

  signature = _urlsafe_b64decode(segments[2])

  # Parse token.
  json_body = _urlsafe_b64decode(segments[1])
  try:
    parsed = simplejson.loads(json_body)
  except:
    raise AppIdentityError('Can\'t parse token: %s' % json_body)

  # Check signature.
  verified = False
  for (keyname, pem) in list(certs.items()):
    verifier = Verifier.from_string(pem, True)
    if (verifier.verify(signed, signature)):
      verified = True
      break
  if not verified:
    raise AppIdentityError('Invalid token signature: %s' % jwt)

  # Check creation timestamp.
  iat = parsed.get('iat')
  if iat is None:
    raise AppIdentityError('No iat field in token: %s' % json_body)
  earliest = iat - CLOCK_SKEW_SECS

  # Check expiration timestamp.
  now = int(time.time())
  exp = parsed.get('exp')
  if exp is None:
    raise AppIdentityError('No exp field in token: %s' % json_body)
  if exp >= now + MAX_TOKEN_LIFETIME_SECS:
    raise AppIdentityError(
      'exp field too far in future: %s' % json_body)
  latest = exp + CLOCK_SKEW_SECS

  if now < earliest:
    raise AppIdentityError('Token used too early, %d < %d: %s' %
      (now, earliest, json_body))
  if now > latest:
    raise AppIdentityError('Token used too late, %d > %d: %s' %
      (now, latest, json_body))

  # Check audience.
  if audience is not None:
    aud = parsed.get('aud')
    if aud is None:
      raise AppIdentityError('No aud field in token: %s' % json_body)
    if aud != audience:
      raise AppIdentityError('Wrong recipient, %s != %s: %s' %
          (aud, audience, json_body))

  return parsed




import mimetypes
from datetime import date, datetime


try:
    import pysftp
except Exception as err:
    pysftp = None

def SEND(task):
    """
    tdata.host = host
    tdata.filename = filename
    # TODO this should be more secure!
    # TODO support ssh keys?
    tdata.username = username
    tdata.password = password
    tdata.data = data

    :param      task:  The task
    :type       task:  { type_description }

    :returns:   { description_of_the_return_value }
    :rtype:     { return_type_description }
    """
    now = datetime.now()
    filename = task.data.filename.format(date=now)
    host = task.data.host
    port = task.data.get("port", 22)
    if ":" in host:
        host, port = host.split(":")
    client = FTPClient(host, port,
        username=task.data.username,
        password=task.data.password,
        default_dir=task.data.path)
    client.put(filename, task.data.data)
    client.disconnect()
    return False

class FTPClient(object):
    def __init__(self, host, port=22, username=None, password=None, default_dir=None):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.cnopts = pysftp.CnOpts()
        self._client = self.connect()
        if default_dir:
            self.chdir(default_dir)

    def __del__(self):
        self.disconnect()

    def connect(self):
        self.cnopts.hostkeys = None
        return pysftp.Connection(host=self.host, 
            port=self.port,
            username=self.username,
            password=self.password,
            cnopts=self.cnopts)

    def disconnect(self):
        if not self._client: return
        self._client.close()

    def chdir(self, path):
        self._client.chdir(path)

    def put(self, file_name, file_data):
        with self._client.open(file_name, 'wb') as f:
            f.write(file_data)



from io import BytesIO
from django.template.loader import get_template
from django.template import Context
from django.http import HttpResponse
from html import escape
from django.conf import settings
from rest.helpers import toString
from rest import settings
import os

try:
    from xhtml2pdf import pisa
except:
    pass

def fetch_resources(uri, rel):
    if settings.get("CDN_DOMAIN", "cdn.") in uri:
        return uri
    path = os.path.join(settings.STATIC_ROOT, uri.replace("/static/", ""))
    return path

def render_to_pdf(context):
    template = get_template(context.template_src)
    html  = template.render(context)
    result = BytesIO()
    pdf = pisa.CreatePDF(html, result, link_callback=fetch_resources)
    return pdf, result.getvalue()

def render_to_pdf_response(context):
    template = get_template(context.template_src)
    html  = template.render(context)
    result = BytesIO()
    pdf = pisa.CreatePDF(html, result, link_callback=fetch_resources)
    if not pdf.err:
        response = HttpResponse(result.getvalue(), content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename={}.pdf'.format(context.filename)
        return response
    return HttpResponse('We had some errors<pre>%s</pre>' % escape(html))


from rest import settings
import requests

TZ_URL = "https://maps.googleapis.com/maps/api/timezone/json"


def getTimeZone(lat, lng):
    rsp = lookup(lat, lng)
    if rsp is not None:
        return rsp.get("timeZoneId", None)
    return None
    

def lookup(lat, lng):
    if settings.GOOGLE_MAPS_KEY is None:
        raise Exception("requires GOOGLE_MAPS_KEY")
    params = dict(location=f"{lat},{lng}", timestamp=0, key=settings.GOOGLE_MAPS_KEY)
    rsp = requests.get(TZ_URL, params)
    print(rsp.text)
    if rsp.status_code == 200:
        rsp = rsp.json()
        return rsp.get("zone", None)
    return None



from django.db import models
from django.conf import settings
from django.http import HttpResponse

from telephony.models import SMS, PhonenumberInfo

from rest import helpers
from rest.views import *
from rest.decorators import *
import importlib

try:
    from twilio.twiml.messaging_response import MessagingResponse
except ImportError:
    pass


@url(r'^sms/msg$')
@url(r'^sms/msg/(?P<pk>\d+)$')
@perm_required("view_logs")
def handle_sms_list(request, pk=None):
    return SMS.on_rest_request(request, pk)


@urlPOST(r'^sms$')
@urlPOST(r'^sms/$')
@login_required
def sendSMS(request):
    me = request.member
    group = request.group
    # for now you can only send msgs to a member
    if "to" in request.DATA:
        to = request.DATA.getlist("to")
        members = Member.objects.filter(pk__in=to)
        message = request.DATA.get("message")
        return restGet(request, SMS.broadcast(members, message, by=me, group=group, transport="sms"))

    if "message" not in request.DATA:
        return restStatus(request, False, error="permission denied")

    member = Member.getFromRequest(request)
    if not member:
        return restStatus(request, False, error="requires valid member")
    phone = member.getProperty("phone")
    if not phone:
        return restStatus(request, False, error="member has no phone number")
    message = request.DATA.get("message")
    # send(endpoint, message, by=None, group=None, to=None, transport="sms", srcpoint=None):
    return restGet(request, SMS.send(phone, message, me, group, member))


@urlPOST(r'^sms/incoming$')
def receiveSMS(request):
    helpers.log_print(request.DATA.asDict())
    SMS.log_incoming(request)
    from_number = request.DATA.get("From")
    handler_name = settings.TELEPHONY_HANDLERS.get(from_number, None)
    if handler_name is not None:
        model = importlib.import_module(handler_name)
        msg = model.on_sms(request)
        if msg is not None:
            resp = MessagingResponse()
            resp.message(msg)
            return HttpResponse(resp.to_xml(), content_type='text/xml')
    resp = MessagingResponse()
    resp.message(settings.TELEPHONY_DEFAULT_SMS_RESPONSE)
    return HttpResponse(resp.to_xml(), content_type='text/xml')


@url(r'^info$')
@url(r'^info/(?P<pk>\d+)$')
@login_required
def handle_info(request, pk=None):
    return PhonenumberInfo.on_rest_request(request, pk)


@url(r'^lookup$')
@login_required
def handle_lookup(request, pk=None):
    number = request.DATA.get(["number", "phone"])
    if number is None:
        return restPermissionDenied(request)
    info = PhonenumberInfo.lookup(number)
    return info.restGet(request)







from ..engines import ffmpeg
from ..render_utils import *
from ..exceptions import *

import random
import tempfile
import logging
import time
from PIL import Image
import os

from django.conf import settings

from rest.log import getLogger
logger = getLogger("medialib", filename="medialib.log")


PresetConfig = {
    'name':		'Video Still',
    'description':	'Still frame from video',
    'stage':	10,
    'default_use':	'still',
    'applicable_kinds': 'VLAS',
    'parameters': {
        'image_container': {
            'description':	'Output image container extension',
            'required':	True,
            'kind':		'C',
            'configurable':	True,
            'choices':	'gif,jpg,png',
            'value':	'jpg',
        },
        'pix_fmt': {
            'description':	'Pixel format',
            'required':	False,
            'kind':		'C',
            'configurable':	True,
            'choices':	'yuv420p,yuyv422,rgb24,bgr24,yuv422p,yuv444p,yuv410p,yuv411p,gray,monow,monob,yuvj420p,yuvj422p,yuvj444p,uyvy422,bgr8,bgr4,bgr4_byte,rgb8,rgb4,rgb4_byte,nv12,nv21,argb,rgba,abgr,bgra,gray16be,gray16le,yuv440p,yuvj440p,yuva420p,rgb48be,rgb48le,rgb565be,rgb565le,rgb555be,rgb555le,bgr565be,bgr565le,bgr555be,bgr555le,yuv420p16le,yuv420p16be,yuv422p16le,yuv422p16be,yuv444p16le,yuv444p16be,rgb444le,rgb444be,bgr444le,bgr444be,bgr48be,bgr48le,yuv420p9be,yuv420p9le,yuv420p10be,yuv420p10le,yuv422p10be,yuv422p10le,yuv444p9be,yuv444p9le,yuv444p10be,yuv444p10le,yuv422p9be,yuv422p9le',
            'value':	'yuvj420p',
        },
        'bucket_size': {
            'description':	'Histogram bucket size',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	32,
        },
        'bucket_max': {
            'description':	'Maximum histogram bucket size before frame is considered solid',
            'required':	True,
            'kind':		'F',
            'configurable':	True,
            'value':	0.7,
        },
        'seek_time': {
            'description':	'Seek from start (sec)',
            'required':	True,
            'kind':		'F',
            'configurable':	True,
            'value':	0.5,
        },
    },
}

def render(item, renditiondef, params):
    original = get_rendition(item, 'original')
    params['no_threads'] = True
    params['frame'] = 1
    
    trycnt = 0
    starttime = time.time()
    maxseek = original.get_meta('duration', 60, float)

    logger.info("starting video_still")
    ok = False
    while not ok:
        logger.info("starting video_still 1")
        try:
            fp = tempfile.NamedTemporaryFile(suffix=('.' + params.get('image_container', 'jpg')), prefix="t_still_")
            dstfile = fp.name

            rotate=original.get_meta('rotate')
            if rotate:
                params['rotate'] = rotate

            ffmpeg.run(dstfile, original.filename, params, ['-f', 'image2'])
            fp.flush()
            fp.seek(0)
            im = Image.open(dstfile)
            im = im.convert("L")
            hist = im.histogram()
            total = im.size[0] * im.size[1]
            bucket_size = params.get('bucket_size', 32)
            i = 0
            j = 0;
            buckets = [0]
            for h in hist:
                if i > bucket_size:
                    i = 0
                    j += 1
                    buckets.append(0)
                buckets[j] += h
                i += 1
            ok = True
            for h in buckets:
                if h > total * params.get('bucket_max', 0.7):
                    ok = False

            if time.time() - starttime > getattr(settings, 'MEDIALIB_STILL_MAX_RUNTIME', 15):
                ok = True

            logger.info("starting video_still 2")
            if ok:
                logger.info("new render for still position")
                new_rendition(item, renditiondef, fp, kind='I', width=im.size[0], height=im.size[1], meta={"still_position": params.get('seek_time', 0)})

            trycnt += 1
            params['seek_time'] = params.get('seek_time', 0) + (0.5 + random.random()) * trycnt

            if params['seek_time'] > maxseek:
                trycnt = 0
                params['seek_time'] = 0
                params['bucket_max'] = params.get('bucket_max', 0.7) + 0.1
        
        except (IOError, SyntaxError, ValueError, EOFError, AttributeError, IndexError, CmdError) as err:
            logger.error("video_still failed")
            try:
                fp.close()
            except:
                pass

            if trycnt == 0 or params.get('bucket_max', 0.7) > 1:
                raise RenderError('%s: %s' % (renditiondef.name, str(err)), stack=err, show_stack=True)
            else:
                logger.error("EXCEPTION:(recovered) %s" % str(err))
                # no suitable image, try again
                trycnt = 0
                params['seek_time'] = 0
                params['bucket_max'] = params.get('bucket_max', 0.7) + 0.1



from rest import settings
import requests
from rest import helpers as rh
from objict import objict

GOOGLE_ID_TOKEN_INFO_URL = 'https://www.googleapis.com/oauth2/v3/tokeninfo'
GOOGLE_ACCESS_TOKEN_OBTAIN_URL = 'https://oauth2.googleapis.com/token'
GOOGLE_USER_INFO_URL = 'https://www.googleapis.com/oauth2/v3/userinfo'


def validateIdToken(id_token) -> bool:
    # Reference: https://developers.google.com/identity/sign-in/web/backend-auth#verify-the-integrity-of-the-id-token
    response = requests.get(
        GOOGLE_ID_TOKEN_INFO_URL,
        params={'id_token': id_token}
    )

    if not response.ok:
        raise Exception('id_token is invalid.')

    audience = response.json()['aud']

    if audience != settings.GOOGLE_OAUTH2_CLIENT_ID:
        raise Exception('Invalid audience.')

    return True


def getAccessToken(code, redirect_uri):
    # Reference: https://developers.google.com/identity/protocols/oauth2/web-server#obtainingaccesstokens
    data = {
        'code': code,
        'client_id': settings.GOOGLE_OAUTH2_CLIENT_ID,
        'client_secret': settings.GOOGLE_OAUTH2_CLIENT_SECRET,
        'redirect_uri': redirect_uri,
        'grant_type': 'authorization_code'
    }

    response = requests.post(GOOGLE_ACCESS_TOKEN_OBTAIN_URL, data=data)
    if not response.ok:
        rh.log_error("google.getAccessToken FAILED", data, "response", response.text)
        return None
    resp = response.json()
    # rh.log_print("getAccessToken", resp)
    return objict.fromdict(resp)


def getUserInfo(access_token):
    """
    Get user information from Google

    :param access_token: Google access token
    :return: User info
    """
    # Reference: https://developers.google.com/identity/protocols/oauth2/web-server#callinganapi
    response = requests.get(
        GOOGLE_USER_INFO_URL,
        params={'access_token': access_token}
    )
    if not response.ok:
        rh.log_error("google.getUserInfo FAILED", response.text)
        return None
    resp = objict.fromdict(response.json())
    # rh.log_print("getUserInfo", resp)
    return resp



from datetime import datetime, date
from objict import objict
from django.db import models as dm
import string

from rest import helpers as rh
from rest import errors as re
from rest.encryption import ENCRYPTER, DECRYPTER


class MetaDataBase(dm.Model):
    class Meta:
        abstract = True

    category = dm.CharField(db_index=True, max_length=32, default=None, null=True, blank=True)

    key = dm.CharField(db_index=True, max_length=80)

    value_format = dm.CharField(max_length=16)
    value = dm.TextField()

    int_value = dm.IntegerField(default=None, null=True, blank=True)
    float_value = dm.IntegerField(default=None, null=True, blank=True)

    def setValue(self, value):
        self.value = "{}".format(value)
        if type(value) is int or self.value in ["0", "1"]:
            if type(value) is int and value > 2147483647:
                self.value_format = "S"
                return
            self.value_format = "I"
            self.int_value = value
        elif type(value) is float:
            self.value_format = "F"
            self.float_value = value
        elif isinstance(value, list):
            self.value_format = "L"
            # self.value = ",".join(value)
        elif isinstance(value, dict):
            self.value_format = "O"
        elif type(value) in [str, str] and len(value) < 9 and value.isdigit():
            self.value_format = "I"
            self.int_value = value
        elif value in ["True", "true", "False", "false"]:
            self.value_format = "I"
            if value in ["True", "true"]:
                self.int_value = 1
            else:
                self.int_value = 0
        elif isinstance(value, bool):
            self.value_format = "I"
            if value:
                self.int_value = 1
            else:
                self.int_value = 0
        else:
            self.value_format = "S"

    def getStrictType(self, field_type):
        if type(self.value) is field_type:
            return self.value
        if field_type in [int, str, float, str]:
            return field_type(self.value)
        elif field_type is bool:
            if self.value_format == 'I':
                return self.int_value != 0
            return self.value in [True, 1, '1', 'y', 'Y', 'Yes', 'true', 'True']
        elif field_type in [date, datetime]:
            return rh.parseDate(self.value)
        return self.value

    def getValue(self, field_type=None):
        if field_type:
            return self.getStrictType(field_type)
        elif self.value_format == 'I':
            return self.int_value
        elif self.value_format == 'F':
            return self.float_value
        elif self.value_format in ["L", "O"] and self.value:
            try:
                return eval(self.value)
            except Exception:
                pass
        return self.value

    def __unicode__(self):
        if self.category:
            return "{}.{}={}".format(self.category, self.key, self.value)
        return "{}={}".format(self.key, self.value)

    def __str__(self):
        if self.category:
            return "{}.{}={}".format(self.category, self.key, self.value)
        return "{}={}".format(self.key, self.value)


class MetaDataModel(object):
    def set_metadata(self, request, values=None):
        # this may get called before the model is saved
        if not self.id:
            self.save()

        if values is None:
            values = request
            request = None

        if isinstance(values, list):
            # bug fix probably for odd browser handling of certain form data
            output = objict()
            for item in values:
                if isinstance(item, dict):
                    output.update(item)
            values = output

        if not isinstance(values, dict):
            raise Exception("invalid metadata: {}".format(values))

        for key, value in list(values.items()):
            cat = None
            if "." in key:
                cat, key = key.split('.')
            self.setProperty(key, value, cat, request=request)

    def metadata(self):
        return self.getProperties()

    def removeProperties(self, category=None):
        # this will remove all properties
        # if category is not it will remove all properties
        self.properties.filter(category=category).delete()

    def getProperties(self, category=None):
        ret = {}
        for p in self.properties.all():
            if p.category:
                props = self.getFieldProps(p.category)
                if props.hidden:
                    continue
                if p.category not in ret or not isinstance(ret.get(p.category, None), dict):
                    ret[p.category] = {}
                props = self.getFieldProps("{}.{}".format(p.category, p.key))
                if props.hidden:
                    continue
                if p.category == "secrets":
                    if p.int_value:
                        ret[p.category][p.key] = "*" * p.int_value
                    else:
                        ret[p.category][p.key] = "******"
                    continue
                ret[p.category][p.key] = p.getValue()
            else:
                props = self.getFieldProps(p.key)
                if props.hidden:
                    continue
                ret[p.key] = p.getValue()
        if category is not None:
            if category in ret:
                return ret[category]
            return {}
        return ret

    def __initFieldProps(self):
        if not hasattr(self, "__field_props"):
            if hasattr(self.RestMeta, "METADATA_FIELD_PROPERTIES"):
                # this provides extra protection for metadata fields
                self.__field_props = self.RestMeta.METADATA_FIELD_PROPERTIES
            else:
                self.__field_props = None

    def getFieldProps(self, key):
        self.__initFieldProps()
        full_key = key
        category = None
        if "." in key:
            category, key = key.split('.')
        props = objict()
        if self.__field_props:
            if category and self.__field_props.get(category, None):
                cat_props = self.__field_props.get(category, None)
                if cat_props:
                    props.notify = cat_props.get("notify", None)
                    props.requires = cat_props.get("requires", None)
                    props.on_change_name = cat_props.get("on_change", None)
                    props.hidden = cat_props.get("hidden", False)
                    if props.on_change_name:
                        props.on_change = getattr(self, props.on_change_name, None)
            field_props = self.__field_props.get(full_key, None)
            if field_props:
                props.notify = field_props.get("notify", props.notify)
                props.requires = field_props.get("requires", None)
                props.hidden = field_props.get("hidden", False)
                on_change_name = field_props.get("on_change", None)
                if on_change_name:
                    on_change = getattr(self, on_change_name, None)
                    if on_change:
                        props.on_change = on_change
        return props

    def checkFieldPerms(self, full_key, props, request=None):
        if not props.requires:
            return True
        if not request or not request.member:
            return False
        if request.member.hasPermission(props.requires) or request.user.is_superuser:
            return True

        # this a unauthorized attempt to change field, log and throw exception
        if props.notify and request.member:
            subject = "permission denied changing protected '{}' field".format(full_key)
            msg = "permission denied changing protected field '{}'\nby user: {}\nfor: {}".format(
                    full_key,
                    request.user.username,
                    self
                )
            request.member.notifyWithPermission(props.notify, subject, msg, email_only=True)
        raise re.PermissionDeniedException(subject)

    def setProperties(self, data, category=None, request=None, using=None):
        for k, v in data.items():
            self.setProperty(k, v, category, request=request, using=using)

    def setProperty(self, key, value, category=None, request=None, using=None, ascii_only=False, encrypted=False):
        # rh.log_print("{}:{} ({})".format(key, value, type(value)))
        if ascii_only and isinstance(value, str):
            value = "".join([x for x in value if x in string.printable])
        on_change = None
        if not using:
            using = getattr(self.RestMeta, "DATABASE", using)
        if not request:
            request = rh.getActiveRequest()
        self.__initFieldProps()

        if isinstance(value, dict):
            return self.setProperties(value, key)
        username = "root"
        if request and request.member:
            username = request.member.username
        prop = None

        if "." in key:
            category, key = key.split('.')
        if category:
            # delete any keys with this category name
            full_key = "{}.{}".format(category, key)
            # this deletes anything with the key that matches the category
            # this works because the category is stored not in key but category field
            # rh.log_print("deleting key={}".format(category))
            self.properties.filter(key=category).delete()
            if category == "secrets":
                encrypted = True
        else:
            full_key = key

        value_len = 0
        if encrypted:
            value_len = len(value)
            value = ENCRYPTER.encrypt(value)

        field_props = self.getFieldProps(full_key)
        if not self.checkFieldPerms(full_key, field_props, request):
            return False

        check_value = "{}".format(value)
        has_changed = False
        prop = self.properties.filter(category=category, key=key).last()
        old_value = None
        if prop:
            # existing property we need to make sure we delete
            old_value = prop.getValue()
            if value is None or value == "":
                prop.delete()
                has_changed = True
            else:
                has_changed = check_value != prop.value
                if not has_changed:
                    return
                prop.setValue(value)
                if encrypted and value_len:
                    prop.int_value = value_len
                prop.save(using=using)
            if field_props.on_change:
                field_props.on_change(key, value, old_value, category)
        elif value is None or value == "":
            # do not create none or empty property
            return False
        else:
            has_changed = True
            PropClass = self.get_fk_model("properties")
            prop = PropClass(parent=self, key=key, category=category)
            prop.setValue(value)
            # rh.log_print(u"saving {}.{}".format(category, key))
            # rh.log_print(u"saving {} : {}".format(full_key, value))
            prop.save(using=using)

        if hasattr(self, "_recordRestChange"):
            self._recordRestChange("metadata.{}".format(full_key), old_value)

        if field_props.notify and request and request.member:
            notify = field_props.get("notify")
            if value:
                value = str(value)
                if len(value) > 5:
                    value = "***"
            msg = "protected field '{}' changed to '{}'\nby user: {}\nfor: {}".format(
                full_key,
                value,
                username,
                self
            )
            request.member.notifyWithPermission(notify, "protected '{}' field changed".format(full_key), msg, email_only=True)
        return has_changed

    def getProperty(self, key, default=None, category=None, field_type=None, decrypted=False):
        try:
            if "." in key:
                category, key = key.split('.')
            value = self.properties.get(category=category, key=key).getValue(field_type)
            if decrypted and value is not None:
                return DECRYPTER.decrypt(value)
            return value
        except Exception:
            pass
        return default

    def setSecretProperty(self, key, value):
        return self.setProperty(key, value, category="secrets", encrypted=True)

    def getSecretProperty(self, key, default=None):
        return self.getProperty(key, default, "secrets", decrypted=True)



from ..engines import ffmpeg
from ..engines import hls
from ..render_utils import *
from ..exceptions import *
from .. import render_utils
from medialib.models import RenditionSegment

from django.utils.http import base36_to_int, int_to_base36

import time
import hashlib
import tempfile
import os
import copy

from django.conf import settings

PresetConfig = {
    'name':		'HLS Video',
    'description':	'HLS x264/AAC video',
    'stage':	100,
    'use':		'video',
    'applicable_kinds': 'VLAS',
    'parameters': {
        'audio_bitrate': {
            'description':	'Audio bitrate (kbps)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	128,
        },
        'audio_channels': {
            'description':	'Audio channels',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	2,
        },
        'audio_freq': {
            'description':	'Audio frequency (hz)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	44100,
        },
        'video_bitrate': {
            'description':	'Video bitrate (kbps)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	512,
        },
        'frame': {
            'description':	'Number of frames to encode',
            'required':	False,
            'kind':		'I',
            'configurable':	True,
        },
        'frame_rate': {
            'description':	'Frame rate',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	30,
        },
        'seek_time': {
            'description':	'Seek from start (sec)',
            'required':	True,
            'kind':		'F',
            'configurable':	True,
        },
        'key_frame': {
            'description':	'Key frame rate',
            'required':	False,
            'kind':		'I',
            'configurable':	True,
            'value':	25,
        },
        'duration': {
            'description':	'Duration',
            'required':	False,
            'kind':		'F',
            'configurable':	True,
        },
        'segment_duration': {
            'description':	'Duration of individual HLS segments (sec)',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	3,
        },
    },
}
PresetConfig['parameters'].update(copy.deepcopy(render_utils.PresetParameters))


def hls_default(rendition):
    path = '/' + int_to_base36(rendition.mediaitem.pk) + '/' + (rendition.rendition_definition and int_to_base36(rendition.rendition_definition.pk) or "H")
    path += hashlib.md5(path + str(int(time.time()))).hexdigest()
    url = rendition.mediaitem.library.default_store() + path + "/"
    rendition.url = url
    rendition.bytes = 0
    return rendition

def subcue(item, renditiondef, params):
    try:
        cue = item.cueparent.all()[0]
    except IndexError:
        raise RenderError("Invalid cueparent")
        
    original = get_rendition(cue.item, 'hls')
    if not original:
        raise RenderError("hls rendition not enable in parent media")
    
    params = calc_size(params, original.width, original.height)
    rendition = hls_default(new_rendition(item, renditiondef, None, kind='V', width=params.get('width', None), height=params.get('height', None)))
    rendition.save()
    sdurr = original.get_meta('segment_duration')
    if sdurr:
        rendition.set_meta('segment_duration', sdurr)

    first = None
    try:
        first = original.segments.get(end__gte = cue.start, start__lte = cue.start)
    except RenderSegment.DoesNotExist:
        for seg in original.segments.all().order_by('start'):
            if seg.start > cue.start:
                first = seg
                break
    if not first:
        raise RenderError("Cue points out of range")
    if first.start >= cue.start - 2:
        cstart = first.start
    else:
        # XXX: break out first segment here
        cstart = first.start
    n = 1
    for seg in original.segments.filter(start__gte = cstart, start__lt = cue.end):
        RenditionSegment(rendition=rendition, segment=n, start=seg.start - cstart, end = seg.end - cstart, duration=seg.duration, bytes=seg.bytes, url=seg.url).save()
        n += 1

def cleanup(data, haserr):
    if haserr:
        ret = data['proc'].wait(nohang=True)
        if ret == None:
            data['proc'].kill()
            data['proc'].wait()
        elif ret == 0:
            ret = -1
    else:
        ret = data['proc'].wait()
    try:
        os.unlink(data['fname'])
    except:
        pass
    if ret:
        raise CmdError(ret, data['proc'].err(), cmd='*hls_segment')

def render(item, renditiondef, params):
    if item.kind == 'S': return subcue(item, renditiondef, params)
    proc = None
    (fp, fname) = tempfile.mkstemp(prefix="t_hls_")
    os.close(fp)
    os.unlink(fname)
    os.mkfifo(fname)

    original = get_rendition(item, 'original')
    params['no_threads'] = False
    params = calc_size(params, original.width, original.height, rotate=original.get_meta('rotate'))
    
    try:
        rendition = hls_default(new_rendition(item, renditiondef, None, width=params.get('width', None), height=params.get('height', None)))
        rendition.save()
        if 'segment_duration' in params:
            rendition.set_meta('segment_duration', params['segment_duration'])

        proc = hls.run(rendition, fname, params)
        ffmpeg_params = params.copy()
        ffmpeg_params['video_codec'] = 'libx264'
        ffmpeg_params['audio_codec'] = 'aac'
        ffmpeg_params['audio_channels'] = 2
        extra_args=('-f', 'mpegts', '-vbsf', 'h264_mp4toannexb', '-async' ,'1' ,'-adrift_threshold' ,'0',)
        if item.kind == 'L':
            ffmpeg.schedule(fname, original.filename, ffmpeg_params, extra_args=extra_args, data={'proc': proc, 'fname': fname}, cleanup=cleanup)
        else:
            ffmpeg.run(fname, original.filename, ffmpeg_params, extra_args=extra_args)
            proc.wait()
            proc = None

    except (IOError, SyntaxError, ValueError, EOFError, AttributeError, IndexError, CmdError) as err:
        raise RenderError('%s: %s' % (renditiondef.name, str(err)), stack=err, show_stack=True)
    finally:
        if item.kind != 'L':
            if proc:
                proc.kill()
            try:
                os.unlink(fname)
            except:
                pass
        
